# -*- coding: utf-8 -*-
import inspect
from abc import abstractmethod, ABC
from collections import defaultdict
from typing import TypeVar, Generic, ClassVar, TYPE_CHECKING, Tuple, Optional, Callable, Any

from ordered_set import OrderedSet

if TYPE_CHECKING:
    from ddb.cache import Cache  # pylint:disable=cyclic-import


class RegistryObject(ABC):
    """
    Object that can be registered in a registry. It should have a unique name.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        The name of the object. It should be unique inside the registry.
        """

    @property
    def description(self) -> str:
        """
        The description of the object.
        """
        return inspect.getdoc(self.__class__)


class DefaultRegistryObject(RegistryObject):
    """
    Default object that can be registered in a registry. It should have a unique name.
    """

    def __init__(self, name: str, description: str = None):
        self._name = name
        self._description = description

    @property
    def name(self) -> str:
        return self._name

    @property
    def description(self) -> str:
        return self._description if self._description else super().description


T = TypeVar('T')  # pylint:disable=invalid-name


class Registry(Generic[T]):
    """
    Container for registered objects. Objects can be retrieved with their unique name.
    """

    def __init__(self, clazz: ClassVar[T], type_name="Object"):
        self._clazz = clazz
        self._type_name = type_name
        self._objects = []
        self._objects_dict = {}
        self._cache = None

    def set_cache(self, cache: 'Cache'):
        """
        Set a cache provider, and register cached entries into registry.
        """
        self._cache = cache
        if self._cache:
            for key in self._cache.keys():
                self._register_cache_value_impl(self._cache.get(key), key)

    def _register_cache_value_impl(self, value, key):
        self.register(value, key)

    def _default_name(self, obj: T):  # pylint:disable=no-self-use
        """
        The name getter for an object.
        """
        if hasattr(obj, "name"):
            return obj.name
        raise ValueError("Name should be provided to register this kind of object")

    def register(self, obj: T, name=None):
        """
        Register an object instance.
        """
        name = name if name else self._default_name(obj)

        if not isinstance(obj, self._clazz):
            raise ValueError(self._type_name + " name \"" +
                             name + "\" should be an instance of " +
                             self._clazz.__name__)

        data = self._register_impl(name, obj)
        return data

    def _register_impl(self, name: str, obj: T):
        if name in self._objects_dict:
            raise ValueError(self._type_name + " name \"" + name + "\" is already registered")

        self._objects_dict[name] = obj
        self._objects.append(obj)

        if self._cache and obj and (name not in self._cache or self._cache.get(name) != obj):
            self._cache.set(name, obj)
            self._cache.flush()

        return obj

    def has(self, name: str, obj: Optional[T] = None) -> bool:
        """
        Check if name is registered.
        """
        ret = name in self._objects_dict
        if ret and obj is not None:
            return self._has_obj_impl(self._objects_dict.get(name), obj)
        return ret

    def _has_obj_impl(self, dict_value, obj) -> bool:  # pylint:disable=no-self-use
        return dict_value == obj

    def unregister(self, name, obj: Optional[T] = None, callback: Optional[Callable[[T], Any]] = None) -> bool:
        """
        Unregister an object instance
        """
        if name not in self._objects_dict:
            raise ValueError(self._type_name + " name \"" + name + "\" is not registered")

        return self._unregister_impl(name, obj, callback)

    def _unregister_impl(self, name, obj=None, callback: Optional[Callable[[T], Any]] = None) -> bool:
        item = self._objects_dict.get(name)
        if obj is None or item == obj:
            self._objects_dict.pop(name)
            self._objects.remove(item)

            if self._cache and name in self._cache:
                self._cache.pop(name)

            if callback:
                callback(item)

            return True
        return False

    def get(self, name: str) -> T:
        """
        Get an object instance from it's unique name.
        """
        if name not in self._objects_dict:
            raise ValueError(self._type_name + " name \"" + name + "\" is not registered")
        return self._objects_dict[name]

    def all(self) -> Tuple[T]:
        """
        Get all object instances.
        """
        return tuple(self._objects)

    def clear(self):
        """
        Remove all object instances.
        """
        self._objects_dict.clear()
        self._objects.clear()
        if self._cache:
            self._cache.clear()

    def close(self):
        """
        Close this registry, closing the underlying cache if defined and then removing all objects instances.
        :return:
        """
        if self._cache:
            self._cache.close()
            self._cache = None
        self.clear()


class RegistryOrderedSet(Registry):
    """
    Container for registered objects. Objects set can be retrieved with their unique name.
    """
    def __init__(self, clazz: ClassVar[T], type_name="Object"):
        super().__init__(clazz, type_name)
        self._objects_dict = defaultdict(OrderedSet)

    def _register_cache_value_impl(self, value, key):
        for obj in value:
            self.register(obj, key)

    def _register_impl(self, name: str, obj: T):
        items = self._objects_dict[name]
        if obj in items:
            raise ValueError("This " + self._type_name + " is already registered for name \"" + name + "\"")
        items.add(obj)
        self._objects.append(obj)

        if self._cache:
            self._cache.set(name, items)
            self._cache.flush()

        return items

    def _unregister_impl(self, name, obj=None, callback: Optional[Callable[[T], Any]] = None) -> bool:
        items = self._objects_dict.get(name)
        ret = False
        if items:
            for item in items:
                if not obj or obj == item:
                    items.remove(item)
                    self._objects.remove(item)
                    ret = True
                    if callback:
                        callback(item)
            if ret:
                if not items:
                    self._objects_dict.pop(name)
                    if self._cache and name in self._cache:
                        self._cache.pop(name)
                else:
                    if self._cache:
                        self._cache.set(name, items)
        return ret

    def _has_obj_impl(self, dict_value: OrderedSet[T], obj) -> bool:
        return obj in dict_value

    def get(self, name: str) -> OrderedSet[T]:  # pylint:disable=useless-super-delegation
        return super().get(name)



import os

import docker
import pytest
import yaml

from ddb.__main__ import load_registered_features, register_default_caches, register_actions_in_event_bus
from ddb.event import events
from ddb.feature import features
from ddb.feature.core import CoreFeature
from ddb.feature.fixuid import FixuidFeature, FixuidDockerComposeAction


class TestFixuidFeature:
    def test_empty_project_without_core(self, project_loader):
        project_loader("empty")

        features.register(FixuidFeature())
        load_registered_features()

        action = FixuidDockerComposeAction()
        action.execute(docker_compose_config={})

    def test_empty_project_with_core(self, project_loader):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(FixuidFeature())
        load_registered_features()

        action = FixuidDockerComposeAction()
        action.execute(docker_compose_config={})

    @pytest.mark.parametrize("project", [
        "from-scratch-empty",
        "from-scratch-with-entrypoint",
        "from-scratch-with-entrypoint-multiple-line",
        "from-scratch-with-entrypoint-string",
        "from-scratch-with-entrypoint-string-no-quotes",
        "from-php-missing-configuration",
        "from-php-empty",
        "from-php-with-entrypoint-only",
        "from-php-with-entrypoint-and-cmd",
        "from-php-user",
        "from-php-root-user"
    ])
    def test_fixuid(self, project_loader, project):
        project_loader(project)
        register_default_caches()

        features.register(FixuidFeature())
        load_registered_features()

        with open('docker-compose.yml', 'r') as config_file:
            config = yaml.load(config_file, yaml.SafeLoader)

        action = FixuidDockerComposeAction()
        action.execute(docker_compose_config=config)

        with open(os.path.join("docker", "Dockerfile.expected"), "r") as f:
            expected = f.read()
        with open(os.path.join("docker", "Dockerfile"), "r") as f:
            content = f.read()
        assert content == expected

        if project in ["from-php-missing-configuration"]:
            assert not os.path.exists(os.path.join("docker", "fixuid.tar.gz"))
        else:
            assert os.path.exists(os.path.join("docker", "fixuid.tar.gz"))

        if "scratch" not in project:
            client = docker.from_env()
            image = client.images.build(path="docker")
            assert image

    def test_from_mysql_fixuid_removed(self, project_loader):
        project_loader("from-mysql-fixuid-removed")
        register_default_caches()

        features.register(FixuidFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        with open('docker-compose.yml', 'r') as config_file:
            config = yaml.load(config_file, yaml.SafeLoader)

        events.docker.docker_compose_config(config)

        with open(os.path.join("docker", "Dockerfile.expected"), "r") as f:
            expected = f.read()
        with open(os.path.join("docker", "Dockerfile"), "r") as f:
            content = f.read()
        assert content == expected

        assert os.path.exists(os.path.join("docker", "fixuid.tar.gz"))

        os.remove(os.path.join("docker", "fixuid.yml"))
        events.file.deleted(os.path.join("docker", "fixuid.yml"))

        with open(os.path.join("docker", "Dockerfile.removed.expected"), "r") as f:
            expected = f.read()
        with open(os.path.join("docker", "Dockerfile"), "r") as f:
            content = f.read()
        assert content == expected

        assert not os.path.exists(os.path.join("docker", "fixuid.tar.gz"))

    def test_no_fixuid(self, project_loader):
        project_loader("no-fixuid")
        register_default_caches()

        features.register(FixuidFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        with open('docker-compose.yml', 'r') as config_file:
            config = yaml.load(config_file, yaml.SafeLoader)

        events.docker.docker_compose_config(config)

        with open(os.path.join("docker", "Dockerfile.expected"), "r") as f:
            expected = f.read()
        with open(os.path.join("docker", "Dockerfile"), "r") as f:
            content = f.read()
        assert content == expected

        assert not os.path.exists(os.path.join("docker", "fixuid.tar.gz"))

    def test_fixuid_manual(self, project_loader):
        project_loader("fixuid-manual")
        register_default_caches()

        features.register(FixuidFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        with open('docker-compose.yml', 'r') as config_file:
            config = yaml.load(config_file, yaml.SafeLoader)

        events.docker.docker_compose_config(config)

        with open(os.path.join("docker", "Dockerfile.expected"), "r") as f:
            expected = f.read()
        with open(os.path.join("docker", "Dockerfile"), "r") as f:
            content = f.read()
        assert content == expected

        assert os.path.exists(os.path.join("docker", "fixuid.tar.gz"))

    def test_fixuid_manual_install(self, project_loader):
        project_loader("fixuid-manual-install")
        register_default_caches()

        features.register(FixuidFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        with open('docker-compose.yml', 'r') as config_file:
            config = yaml.load(config_file, yaml.SafeLoader)

        events.docker.docker_compose_config(config)

        with open(os.path.join("docker", "Dockerfile.expected"), "r") as f:
            expected = f.read()
        with open(os.path.join("docker", "Dockerfile"), "r") as f:
            content = f.read()
        assert content == expected

        assert os.path.exists(os.path.join("docker", "fixuid.tar.gz"))

    def test_fixuid_manual_entrypoint(self, project_loader):
        project_loader("fixuid-manual-entrypoint")
        register_default_caches()

        features.register(FixuidFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        with open('docker-compose.yml', 'r') as config_file:
            config = yaml.load(config_file, yaml.SafeLoader)

        events.docker.docker_compose_config(config)

        with open(os.path.join("docker", "Dockerfile.expected"), "r") as f:
            expected = f.read()
        with open(os.path.join("docker", "Dockerfile"), "r") as f:
            content = f.read()
        assert content == expected

        assert os.path.exists(os.path.join("docker", "fixuid.tar.gz"))



# -*- coding: utf-8 -*-
import base64
import json
import os
import shlex
from typing import Iterable

import zgitignore
from dictdiffer import diff

from .integrations import ShellIntegration
from ...action import Action, actions
from ...action.action import EventBinding
from ...binary import Binary, binaries
from ...binary.binary import DefaultBinary
from ...config import config
from ...config.flatten import to_environ
from ...context import context
from ...event import events
from ...utils.file import SingleTemporaryFile

_env_environ_backup = config.env_prefix + "_SHELL_ENVIRON_BACKUP"


def get_shims_path(global_: bool = False):
    """
    Get local or global shims path.
    :param global_:
    :return:
    """
    directories = config.data.get('shell.path.directories')

    if global_:
        shim_directory_prefix = config.paths.home
        forbidden_shim_directory_prefix = config.paths.project_home
    else:
        shim_directory_prefix = config.paths.project_home
        forbidden_shim_directory_prefix = None

    shim_directory = None
    for directory in directories:
        if not os.path.isabs(directory):
            directory = os.path.join(config.paths.project_home, directory)
        if directory.startswith(shim_directory_prefix) and \
                (not forbidden_shim_directory_prefix or
                 not directory.startswith(forbidden_shim_directory_prefix)):
            shim_directory = directory
            break

    return shim_directory


def encode_environ_backup(environ_backup: dict) -> str:
    """
    Encode environ backup into a safe string.
    """
    return base64.b64encode(json.dumps(environ_backup).encode("utf-8")).decode("utf-8")


def decode_environ_backup(encoded_environ_backup: str) -> dict:
    """
    Decode environ backup from safe string.
    """
    return json.loads(base64.b64decode(encoded_environ_backup).decode("utf-8"))


def apply_diff_to_shell(shell: ShellIntegration,
                        source_environment: dict,
                        target_environment: dict,
                        envignore=None) -> Iterable[str]:
    """
    Compute and apply environment diff for given shell, returning a list of shell instruction to run.
    """
    variables = []

    for (action, source, dest_items) in diff(source_environment, target_environment):
        if action == 'add':
            for (key, value) in dest_items:
                variables.append((key, value))
        if action == 'change':
            variables.append((source, dest_items[1]))
        if action == 'remove':
            for (key, value) in dest_items:
                variables.append((key, None))

    envignore_helper = None
    if envignore:
        envignore_helper = zgitignore.ZgitIgnore(envignore)

    for key, value in sorted(variables, key=lambda variable: variable[0]):
        if not envignore_helper or not envignore_helper.is_ignored(key):
            if value is None:
                yield from shell.remove_environment_variable(key)
            else:
                yield from shell.set_environment_variable(key, value)


def add_to_system_path(shell: ShellIntegration, paths: Iterable[str]) -> Iterable[str]:
    """
    Add given path to system PATH environment variable
    """
    system_path = os.environ.get('PATH', '')
    for path in paths:
        if config.data.get('shell.path.prepend'):
            system_path = path + os.pathsep + system_path
        else:
            system_path = system_path + os.pathsep + path

    yield from shell.set_environment_variable('PATH', system_path)


def remove_from_system_path(shell: ShellIntegration, paths: Iterable[str]) -> Iterable[str]:
    """
    Remove given path to system PATH environment variable
    """
    system_path = os.environ.get('PATH', '')
    for path in paths:
        if config.data.get('shell.path.prepend'):
            system_path = system_path.replace(path + os.pathsep, "")
        else:
            system_path = system_path.replace(os.pathsep + path, "")

    yield from shell.set_environment_variable('PATH', system_path)


class CreateBinaryShim(Action):
    """
    Create binary shim for each generated binary
    """

    def __init__(self, shell: ShellIntegration):
        super().__init__()
        self.shell = shell

    @property
    def event_bindings(self):
        return (events.binary.registered,
                events.binary.found,
                EventBinding(events.binary.unregistered, call=self.remove))

    @property
    def name(self) -> str:
        return "shell:" + self.shell.name + ":create-binary-shim"

    @property
    def description(self) -> str:
        return super().description + " for " + self.shell.description

    @property
    def disabled(self) -> bool:
        return config.data.get('shell.shell') != self.shell.name

    def remove(self, binary: Binary):
        """
        Remove binary shim
        """
        directories = config.data.get('shell.path.directories')
        if not binaries.has(binary.name):
            # Remove shim only if no binary are remaining for this name
            removed = self.shell.remove_binary_shim(directories[0], binary.name)
            if removed:
                events.file.deleted(removed)

    def execute(self, binary: Binary):
        """
        Create binary shim
        """
        shims_path = get_shims_path(binary.global_)
        written, shim = self.shell.create_binary_shim(shims_path, binary.name, binary.global_)
        if written:
            context.log.success("Shim created: %s", shim)
        else:
            context.log.notice("Shim exists: %s", shim)

        if written or config.eject:
            events.file.generated(source=None, target=shim)


class CreateAliasShim(Action):
    """
    Create alias shim for each alias
    """

    def __init__(self, shell: ShellIntegration):
        super().__init__()
        self.shell = shell

    @property
    def event_bindings(self):
        return events.phase.configure

    @property
    def name(self) -> str:
        return "shell:" + self.shell.name + ":create-alias-shim"

    @property
    def description(self) -> str:
        return super().description + " for " + self.shell.description

    @property
    def disabled(self) -> bool:
        return config.data.get('shell.shell') != self.shell.name

    def execute(self):
        """
        Create binary shim
        """
        aliases = config.data.get('shell.aliases')
        global_aliases = frozenset(config.data.get('shell.global_aliases'))
        for alias in aliases:
            global_alias = alias in global_aliases

            shim_directory = get_shims_path(global_alias)
            if not shim_directory:
                continue

            binary = DefaultBinary(alias, shlex.split(aliases[alias]))
            written, shim = self.shell.create_alias_binary_shim(shim_directory, binary)
            if written:
                context.log.success("Shim created: %s", shim)
            else:
                context.log.notice("Shim exists: %s", shim)

            if not global_alias:
                if written or config.eject:
                    events.file.generated(source=None, target=shim)


class ActivateAction(Action):
    """
    Generates activation script
    """

    def __init__(self, shell: ShellIntegration):
        super().__init__()
        self.shell = shell

    @property
    def event_bindings(self):
        return "phase:activate"

    @property
    def name(self) -> str:
        return "shell:" + self.shell.name + ":activate"

    @property
    def description(self) -> str:
        return super().description + " for " + self.shell.description

    @property
    def disabled(self) -> bool:
        return config.data.get('shell.shell') != self.shell.name

    @property
    def order(self) -> int:
        return -256

    def _deactivate(self):
        deactivate_action = actions.get("shell:" + self.shell.name + ":deactivate")
        deactivate_action.execute()

    def execute(self):
        """
        Execute action
        """
        try:
            check_activated('force' not in config.args or not config.args.force)
            if config.args.force:
                self._deactivate()
            else:
                raise CheckIsActivatedException("project is already activated.")
        except CheckIsNotActivatedException:
            pass

        initial_environ = dict(os.environ.items())
        config_environ = to_environ(config.data, config.env_prefix)
        config_environ.update(config.env_additions)

        to_encode_environ = dict(initial_environ)
        self.shell.before_environ_backup(to_encode_environ)
        os.environ.update(config_environ)
        os.environ[_env_environ_backup] = encode_environ_backup(to_encode_environ)
        os.environ[config.env_prefix + '_PROJECT_HOME'] = config.paths.project_home

        with SingleTemporaryFile("ddb", "activate",
                                 mode='w',
                                 prefix="",
                                 suffix="." + self.shell.name,
                                 **self.shell.temporary_file_kwargs) as file:
            script_filepath = file.name
            file.writelines('\n'.join(self.shell.header()))
            file.write('\n')

            file.writelines('\n'.join(apply_diff_to_shell(
                self.shell,
                initial_environ,
                os.environ,
                config.data.get('shell.envignore'))))
            file.write('\n')

            path_directories = config.data.get('shell.path.directories')
            if path_directories:
                path_additions = map(
                    lambda path_addition: os.path.normpath(os.path.join(config.paths.project_home, path_addition)),
                    path_directories)
                file.writelines('\n'.join(add_to_system_path(self.shell, path_additions)))
                file.write('\n')

            file.writelines('\n'.join(self.shell.footer()))
            file.write('\n')

        for ins in self.shell.evaluate_script(script_filepath):
            print(ins)


class DeactivateAction(Action):
    """
    Generations deactivation script
    """

    def __init__(self, shell: ShellIntegration):
        super().__init__()
        self.shell = shell

    @property
    def event_bindings(self):
        return "phase:deactivate"

    @property
    def name(self) -> str:
        return "shell:" + self.shell.name + ":deactivate"

    @property
    def description(self) -> str:
        return super().description + " for " + self.shell.description

    @property
    def disabled(self) -> bool:
        return config.data.get('shell.shell') != self.shell.name

    def execute(self):
        """
        Execute action
        """
        try:
            check_activated('force' not in config.args or not config.args.force)
        except CheckAnotherProjectActivatedException:
            pass

        if os.environ.get(_env_environ_backup):
            environ_backup = decode_environ_backup(os.environ[_env_environ_backup])

            with SingleTemporaryFile("ddb", "deactivate",
                                     mode='w',
                                     prefix="",
                                     suffix="." + self.shell.name,
                                     **self.shell.temporary_file_kwargs) as file:
                script_filepath = file.name
                file.writelines('\n'.join(self.shell.header()))
                file.write('\n')

                file.writelines('\n'.join(apply_diff_to_shell(
                    self.shell,
                    os.environ,
                    environ_backup,
                    config.data.get('shell.envignore'))))
                file.write('\n')

                file.writelines('\n'.join(self.shell.footer()))
                file.write('\n')

            for ins in self.shell.evaluate_script(script_filepath):
                print(ins)


class CheckActivatedException(Exception):
    """
    Exception for activated check
    """


class CheckIsActivatedException(CheckActivatedException):
    """
    Exception for activated check
    """


class CheckIsNotActivatedException(CheckActivatedException):
    """
    Exception for activated check
    """


class CheckAnotherProjectActivatedException(CheckActivatedException):
    """
    Exception for activated check
    """


def check_activated(do_raise=False, do_log=False):
    """
    Check if project is activated in current shell.
    :return:
    """
    project_home_key = config.env_prefix + '_PROJECT_HOME'

    if project_home_key in os.environ:
        if os.environ[project_home_key] == config.paths.project_home:
            if do_log:
                context.log.info("Project is activated.")
            return True
        try:
            raise CheckAnotherProjectActivatedException(
                "Another project is already activated (%s)" % os.environ[project_home_key])
        except CheckAnotherProjectActivatedException as exc:
            if do_log:
                context.log.error(str(exc))
            if do_raise:
                raise exc
        return False

    try:
        raise CheckIsNotActivatedException("Project is not activated")
    except CheckIsNotActivatedException as exc:
        if do_log:
            context.log.error(str(exc))
        if do_raise:
            raise exc
        return False


class CheckActivatedAction(Action):
    """
    Check if project is activated in current shell
    """

    def __init__(self, shell: ShellIntegration):
        super().__init__()
        self.shell = shell

    @property
    def event_bindings(self):
        return "phase:check-activated"

    @property
    def name(self) -> str:
        return "shell:" + self.shell.name + ":check-activated"

    @property
    def description(self) -> str:
        return super().description + " for " + self.shell.description

    @property
    def disabled(self) -> bool:
        return config.data.get('shell.shell') != self.shell.name

    def execute(self):  # pylint:disable=no-self-use
        """
        Execute action
        """
        try:
            check_activated(True, True)
        except CheckActivatedException as exc:
            context.exceptions.append(exc)


class CommandAction(Action):
    """
    Display a requested command for it to be evaluated by the shell
    """

    def __init__(self, shell: ShellIntegration):
        super().__init__()
        self.shell = shell

    @property
    def event_bindings(self):
        return events.run.command

    @property
    def name(self) -> str:
        return "shell:" + self.shell.name + ":command"

    @property
    def description(self) -> str:
        return super().description + " for " + self.shell.description

    @property
    def disabled(self) -> bool:
        return config.data.get('shell.shell') != self.shell.name

    def execute(self, command: Iterable[str], system_path: bool = False):
        """
        Execute action
        """
        if not system_path:
            print(self.shell.generate_cmdline(command))
            return

        path_additions = None

        with SingleTemporaryFile("ddb", "run",
                                 mode='w',
                                 prefix="",
                                 suffix="." + self.shell.name,
                                 **self.shell.temporary_file_kwargs) as file:
            script_filepath = file.name
            if system_path:
                path_directories = config.data.get('shell.path.directories')
                if path_directories:
                    path_additions = map(
                        lambda path_addition: os.path.normpath(os.path.join(config.paths.project_home, path_addition)),
                        path_directories)
                    for instruction in remove_from_system_path(self.shell, path_additions):
                        print(instruction, file=file)
                    print(file=file)

            print(self.shell.generate_cmdline(command), end="", file=file)

            if path_additions:
                or_generated = False
                for instruction in add_to_system_path(self.shell, path_additions):
                    if not or_generated:
                        print(self.shell.generate_or_operator(True), file=file)
                    else:
                        print(self.shell.generate_and_operator(True), file=file)
                    print(instruction, end="", file=file)

            print(file=file)

        for ins in self.shell.evaluate_script(script_filepath):
            print(ins)



from ddb.__main__ import register_default_caches
from ddb.cache import global_cache, project_cache


def test_global_cache():
    register_default_caches()
    assert global_cache() is not None


def test_project_cache():
    register_default_caches()
    assert project_cache() is not None



# -*- coding: utf-8 -*-
from typing import ClassVar, Iterable

from ddb.action import Action
from ddb.feature import Feature
from .actions import PermissionsAction
from .schema import PermissionsSchema


class PermissionsFeature(Feature):
    """
    Enforce permissions on files from defined patterns.
    """

    @property
    def name(self) -> str:
        return "permissions"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core"]

    @property
    def schema(self) -> ClassVar[PermissionsSchema]:
        return PermissionsSchema

    @property
    def actions(self) -> Iterable[Action]:
        return (
            PermissionsAction(),
        )



# -*- coding: utf-8 -*-
import os
import pathlib
import re
from typing import Iterable, ClassVar

import netifaces
from dotty_dict import Dotty

from .actions import EmitDockerComposeConfigAction, DockerComposeBinaryAction, LocalVolumesAction, \
    DockerDisplayInfoAction
from .schema import DockerSchema
from ..feature import Feature, FeatureConfigurationAutoConfigureError
from ..schema import FeatureSchema
from ...action import Action
from ...config import config


class DockerFeature(Feature):
    """
    Docker and docker-compose support.
    """

    @property
    def name(self) -> str:
        return "docker"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core"]

    @property
    def schema(self) -> ClassVar[FeatureSchema]:
        return DockerSchema

    @property
    def actions(self) -> Iterable[Action]:
        return (
            EmitDockerComposeConfigAction(),
            DockerComposeBinaryAction(),
            LocalVolumesAction(),
            DockerDisplayInfoAction()
        )

    def _configure_defaults(self, feature_config: Dotty):
        self._configure_defaults_ip(feature_config)
        self._configure_defaults_user_from_name_and_group(feature_config)
        self._configure_defaults_user(feature_config)
        self._configure_defaults_path_mapping(feature_config)

    def _configure_defaults_ip(self, feature_config):
        ip_address = feature_config.get('ip')
        if not ip_address:
            docker_host = os.environ.get('DOCKER_HOST')
            if docker_host:
                ip_match = re.match(r"(?:.*?)://(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}).*", docker_host)
                if ip_match:
                    ip_address = ip_match.group(1)

        try:
            if not ip_address:
                interface = feature_config.get('interface')
                try:
                    docker_if = netifaces.ifaddresses(interface)
                except ValueError as err:
                    raise FeatureConfigurationAutoConfigureError(self, 'ip',
                                                                 "Invalid network interface: " + interface) from err
                if docker_if and netifaces.AF_INET in docker_if:
                    docker_af_inet = docker_if[netifaces.AF_INET][0]
                    ip_address = docker_af_inet['addr']
                else:
                    raise FeatureConfigurationAutoConfigureError(self, 'ip',
                                                                 "Can't get ip address "
                                                                 "from network interface configuration: " + interface)
        except FeatureConfigurationAutoConfigureError as error:
            if os.path.exists('/var/run/docker.sock'):
                ip_address = '127.0.0.1'
            else:
                raise error

        feature_config['ip'] = ip_address

    @staticmethod
    def _configure_defaults_user_from_name_and_group(feature_config):
        uid = feature_config.get('user.uid')
        gid = feature_config.get('user.gid')

        if uid is None or gid is None:
            name = feature_config.get('user.name')
            if name:
                try:
                    import pwd  # pylint:disable=import-outside-toplevel
                    struct_passwd = pwd.getpwnam(name)
                    if uid is None:
                        uid = struct_passwd.pw_uid
                        feature_config['user.uid'] = uid
                    if gid is None:
                        gid = struct_passwd.pw_gid
                        feature_config['user.gid'] = gid
                except ImportError:
                    pass
                except KeyError:
                    pass

            group = feature_config.get('user.group')
            if group:
                try:
                    import grp  # pylint:disable=import-outside-toplevel
                    struct_group = grp.getgrnam(group)
                    gid = struct_group.gr_id
                    feature_config['user.gid'] = gid
                except ImportError:
                    pass
                except KeyError:
                    pass

    @staticmethod
    def _configure_defaults_user(feature_config):
        uid = feature_config.get('user.uid')
        if uid is None:
            try:
                uid = os.getuid()  # pylint:disable=no-member
            except AttributeError:
                uid = 1000
            feature_config['user.uid'] = uid

        gid = feature_config.get('user.gid')
        if gid is None:
            try:
                gid = os.getgid()  # pylint:disable=no-member
            except AttributeError:
                gid = 1000
            feature_config['user.gid'] = gid

    @staticmethod
    def _configure_defaults_path_mapping(feature_config):
        """
        On windows, this generates a default path mapping matching docker-compose behavior when
        COMPOSE_CONVERT_WINDOWS_PATHS=1 is enabled.

        Drive letter should be lowercased to have the same behavior

        https://github.com/docker/compose/blob/f1059d75edf76e8856469108997c15bb46a41777/compose/config/types.py#L123-L132
        """
        path_mapping = feature_config.get('path_mapping')
        if path_mapping is None:
            path_mapping = {}
            if config.data.get('core.os') == 'nt':
                raw = config.data.get('core.path.project_home')
                mapped = re.sub(r"^([a-zA-Z]):", r"/\1", raw)
                mapped = pathlib.Path(mapped).as_posix()
                mapped = re.sub(r"(\/)(.)(\/.*)", lambda x: x.group(1) + x.group(2).lower() + x.group(3), mapped)
                path_mapping[raw] = mapped
            feature_config['path_mapping'] = path_mapping



import os

from ddb.__main__ import load_registered_features, register_actions_in_event_bus
from ddb.config import migrations, config
from ddb.config.migrations import PropertyMigration
from ddb.feature import features
from ddb.feature.core import CoreFeature
from ddb.feature.file import FileFeature, FileWalkAction
from ddb.feature.ytt import YttFeature


class TestYttAction:
    def test_empty_project_without_core(self, project_loader):
        project_loader("empty")

        features.register(FileFeature())
        features.register(YttFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

    def test_empty_project_with_core(self, project_loader):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(YttFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

    def test_plain(self, project_loader):
        project_loader("plain")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(YttFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('yaml.yml')
        with open('yaml.yml', 'r') as f:
            rendered = f.read()

        with open('yaml.expected.yml', 'r') as f:
            expected = f.read()

        assert rendered == expected

    def test_config_variables(self, project_loader):
        project_loader("config_variables")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(YttFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('variables.yaml')
        with open('variables.yaml', 'r') as f:
            rendered = f.read()

        with open('variables.expected.yaml', 'r') as f:
            expected = f.read()

        assert rendered == expected

    def test_ignore_invalid_extension(self, project_loader):
        project_loader("ignore_invalid_extension")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(YttFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert not os.path.exists('yaml.txt')
        assert not os.path.exists('yaml.yml')

    def test_depends_suffixes(self, project_loader):
        project_loader("depends_suffixes")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(YttFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('variables.yaml')
        with open('variables.yaml', 'r') as f:
            rendered = f.read()

        with open('variables.expected.yaml', 'r') as f:
            expected = f.read()

        assert rendered == expected


class TestYttAutofix:
    def teardown_method(self, test_method):
        migrations.set_history()

    def test_autofix_variables_only(self, project_loader):
        project_loader("autofix_variables_only")

        config.args.autofix = True

        history = (
            PropertyMigration("old_property",
                              "new_property", since="v1.1.0"),
            PropertyMigration("some.deep.old.property",
                              "some.another.new.property", since="v1.1.0"),
        )

        migrations.set_history(history)

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(YttFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('variables.yaml')
        with open('variables.yaml', 'r') as f:
            rendered = f.read()

        with open('variables.expected.yaml', 'r') as f:
            expected = f.read()

        assert expected == rendered

        with open('variables.ytt.yaml', 'r') as f:
            source = f.read()

        with open('variables.yaml.autofix', 'r') as f:
            fixed = f.read()

        assert source == fixed



# -*- coding: utf-8 -*-
import hashlib
import os
import re
from typing import ClassVar, Iterable

from dotty_dict import Dotty
from slugify import slugify

from ddb.action import Action
from ddb.feature import Feature
from .actions import JsonnetAction
from .schema import JsonnetSchema
from ...config import config
from ...utils.file import TemplateFinder


class JsonnetFeature(Feature):
    """
    Render template files with Jsonnet (https://jsonnet.org).
    """

    @property
    def name(self) -> str:
        return "jsonnet"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core", "file", "docker[optional]", "version[optional]"]

    @property
    def schema(self) -> ClassVar[JsonnetSchema]:
        return JsonnetSchema

    @property
    def actions(self) -> Iterable[Action]:
        return (
            JsonnetAction(),
        )

    def _configure_defaults(self, feature_config: Dotty):
        self._configure_defaults_includes(feature_config)
        self._configure_defaults_user_from_name_and_group(feature_config)
        self._configure_defaults_user(feature_config)
        self._configure_defaults_user_maps(feature_config)
        self._configure_defaults_xdebug(feature_config)
        self._configure_defaults_port_prefix(feature_config)
        self._configure_defaults_compose_project_name(feature_config)
        self._configure_defaults_build_image_tag(feature_config)
        self._configure_defaults_restart_policy(feature_config)
        self._configure_defaults_https(feature_config)

    @staticmethod
    def _configure_defaults_includes(feature_config):
        includes = feature_config.get("includes")
        if includes is None:
            includes = TemplateFinder.build_default_includes_from_suffixes(
                feature_config["suffixes"],
                feature_config["extensions"]
            )
            feature_config["includes"] = includes

    @staticmethod
    def _configure_defaults_user_from_name_and_group(feature_config):
        uid = feature_config.get('docker.user.uid', config.data.get('docker.user.uid'))
        gid = feature_config.get('docker.user.gid', config.data.get('docker.user.gid'))

        if uid is None or gid is None:
            name = feature_config.get('docker.user.name')
            if name:
                try:
                    import pwd  # pylint:disable=import-outside-toplevel
                    struct_passwd = pwd.getpwnam(name)
                    if uid is None:
                        uid = struct_passwd.pw_uid
                        feature_config['docker.user.uid'] = uid
                    if gid is None:
                        gid = struct_passwd.pw_gid
                        feature_config['docker.user.gid'] = gid
                except ImportError:
                    pass
                except KeyError:
                    pass

            group = feature_config.get('docker.user.group')
            if group:
                try:
                    import grp  # pylint:disable=import-outside-toplevel
                    struct_group = grp.getgrnam(group)
                    gid = struct_group.gr_id
                    feature_config['docker.user.gid'] = gid
                except ImportError:
                    pass
                except KeyError:
                    pass

    @staticmethod
    def _configure_defaults_user_maps(feature_config):
        name_to_uid = feature_config.get('docker.user.name_to_uid')
        group_to_gid = feature_config.get('docker.user.group_to_gid')

        try:
            import pwd  # pylint:disable=import-outside-toplevel
            struct_passwds = pwd.getpwall()
            for struct_passwd in struct_passwds:
                name_to_uid[struct_passwd.pw_name] = struct_passwd.pw_uid
                group_to_gid[struct_passwd.pw_name] = struct_passwd.pw_gid
            feature_config['docker.user.name_to_uid'] = name_to_uid
            feature_config['docker.user.group_to_gid'] = group_to_gid
        except ImportError:
            pass

        try:
            import grp  # pylint:disable=import-outside-toplevel
            struct_groups = grp.getgrall()
            for struct_group in struct_groups:
                group_to_gid[struct_group.gr_name] = struct_group.gr_gid
            feature_config['docker.user.group_to_gid'] = group_to_gid
        except ImportError:
            pass

    @staticmethod
    def _configure_defaults_user(feature_config):
        uid = feature_config.get('docker.user.uid', config.data.get('docker.user.uid'))
        if uid is None:
            try:
                uid = os.getuid()  # pylint:disable=no-member
            except AttributeError:
                uid = 1000
            feature_config['docker.user.uid'] = uid

        gid = feature_config.get('docker.user.gid', config.data.get('docker.user.gid'))
        if gid is None:
            try:
                gid = os.getgid()  # pylint:disable=no-member
            except AttributeError:
                gid = 1000
            feature_config['docker.user.gid'] = gid

    @staticmethod
    def _configure_defaults_xdebug(feature_config):
        if feature_config.get('docker.xdebug.host') is None:
            feature_config['docker.xdebug.host'] = config.data.get('docker.ip')

        if feature_config.get('docker.xdebug.session') is None:
            feature_config['docker.xdebug.session'] = config.data.get('core.project.name')

        if not feature_config.get('docker.xdebug.disabled'):
            if feature_config['docker.xdebug.host'] is None:
                feature_config['docker.xdebug.disabled'] = True

            if 'core.env.current' in config.data and 'core.env.available' in config.data:
                feature_config['docker.xdebug.disabled'] = config.data['core.env.current'] != \
                                                           config.data['core.env.available'][-1]
            else:
                feature_config['docker.xdebug.disabled'] = False

    @staticmethod
    def _configure_defaults_port_prefix(feature_config):
        port_prefix = feature_config.get('docker.expose.port_prefix')
        if port_prefix is None:
            project_name = config.data.get('core.project.name')
            if project_name:
                port_prefix = int(hashlib.sha1(project_name.encode('utf-8')).hexdigest(), 16) % 655
                feature_config['docker.expose.port_prefix'] = port_prefix

    @staticmethod
    def _configure_defaults_compose_project_name(feature_config):
        """
        See https://github.com/docker/compose/blob/440c94ea7a7e62b3de50722120ca34c4e818205a/compose/cli/command.py#L181
        """
        compose_project_name = feature_config.get('docker.compose.project_name')
        if compose_project_name is None:
            compose_project_name = feature_config.get('core.project.name')

        def normalize_name(name):
            return re.sub(r'[^-_a-z0-9]', '', name.lower()).replace('_', '-')

        if not compose_project_name:
            compose_project_name = os.path.basename(os.path.abspath(config.paths.project_home))

        compose_project_name = normalize_name(compose_project_name)
        feature_config['docker.compose.project_name'] = compose_project_name
        config.env_additions['COMPOSE_PROJECT_NAME'] = compose_project_name

        compose_network_name = feature_config.get('docker.compose.network_name')
        if not compose_network_name:
            compose_network_name = compose_project_name + "-default"

        compose_network_name = normalize_name(compose_network_name)
        feature_config['docker.compose.network_name'] = compose_network_name
        config.env_additions['COMPOSE_NETWORK_NAME'] = compose_network_name

    @staticmethod
    def _configure_defaults_build_image_tag(feature_config):
        build_image_tag_from = feature_config.get('docker.build.image_tag_from')
        build_image_tag = feature_config.get('docker.builder.image_tag')
        if build_image_tag is None and build_image_tag_from:
            if isinstance(build_image_tag_from, str):
                tag = config.data.get(build_image_tag_from)
                if tag:
                    build_image_tag = tag
            else:
                # Use tag if we are exactly on a tag, branch otherwise
                tag = config.data.get('version.tag')
                version = config.data.get('version.version')
                branch = config.data.get('version.branch')
                if tag and version and tag == version:
                    build_image_tag = tag
                elif branch:
                    build_image_tag = branch
            if build_image_tag:
                feature_config['docker.build.image_tag'] = slugify(build_image_tag,
                                                                   regex_pattern=r'[^\w][^\w.-]{0,127}')

    @staticmethod
    def _configure_defaults_restart_policy(feature_config):
        restart_policy = feature_config.get('docker.service.restart')
        if restart_policy is None:
            if config.data.get("core.env.current") == "dev":
                feature_config['docker.service.restart'] = 'no'
            else:
                feature_config['docker.service.restart'] = 'unless-stopped'

    @staticmethod
    def _configure_defaults_https(feature_config):
        https = feature_config.get('docker.virtualhost.https')
        redirect_to_https = feature_config.get('docker.virtualhost.redirect_to_https')

        if redirect_to_https and not https:
            feature_config['docker.virtualhost.redirect_to_https'] = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import inspect
import logging
import os
import sys
import threading
from argparse import ArgumentParser, Namespace
from gettext import gettext as _
from importlib import import_module
from typing import Optional, Sequence, Iterable, Callable, Union, List

import verboselogs
from colorlog import default_log_colors, ColoredFormatter

from ddb.action import actions
from ddb.action.action import EventBinding, Action, WatchSupport
from ddb.action.runner import ExpectedError
from ddb.action.runnerfactory import action_event_binding_runner_factory
from ddb.binary import binaries
from ddb.cache import caches, global_cache_name, requests_cache_name, \
    project_binary_cache_name, register_global_cache, project_cache_name
from ddb.cache import register_project_cache
from ddb.command import commands
from ddb.command.command import execute_command, Command
from ddb.config import config
from ddb.context import context
from ddb.event import bus, events
from ddb.feature import features, Feature
from ddb.feature.bootstrap import reset_available_features, append_available_feature, \
    load_bootstrap_config, bootstrap_register_features
from ddb.feature.core import ConfigureSecondPassException
from ddb.phase import phases
from ddb.registry import Registry, RegistryObject
from ddb.service import services

_watch_started_event = threading.Event()
_watch_stop_event = threading.Event()


def load_plugins():
    """Load plugins"""
    for directory in [d for d in config.paths if d]:
        extension_directory = os.path.join(directory, ".ddb")
        if os.path.exists(extension_directory):
            sys.path.append(os.path.abspath(extension_directory))
            for dirpath, _, files in os.walk(extension_directory):
                module_name = os.path.relpath(dirpath, extension_directory).replace(os.sep, ".")
                if module_name == '.':
                    module_name = None

                for file in files:
                    if file == "__init__.py":
                        _load_plugins_module(module_name)
                    elif file.endswith("py"):
                        file_module_name = os.path.splitext(file)[0]
                        _load_plugins_module(
                            file_module_name if not module_name else '.'.join((module_name, file_module_name)))


def _load_plugins_module(module_name):
    if not module_name:
        return
    module = import_module(module_name)
    for _, clazz in inspect.getmembers(module):
        if inspect.isclass(clazz) and \
                clazz.__module__ == module.__name__ and \
                not clazz.__name__.startswith("_") and \
                issubclass(clazz, Feature):
            feature = clazz()
            append_available_feature(feature)


def prepare_project_home():
    """
    Set working directory to project home.
    """
    if config.paths.project_home:
        config.cwd = os.getcwd()
        os.chdir(config.paths.project_home)


def register_default_caches():
    """
    Register default caches.
    """
    register_project_cache(project_cache_name)
    binaries.set_cache(register_project_cache(project_binary_cache_name))
    register_global_cache(global_cache_name)
    register_global_cache(requests_cache_name)


def register_objects(features_list: Iterable[Feature],
                     objects_getter: Callable[[Feature], Iterable[RegistryObject]],
                     registry: Registry[RegistryObject]):
    """
    Register objects from features inside registry.
    """
    all_objects = []

    for feature in features_list:
        objects = objects_getter(feature)
        all_objects.extend(objects)

    for obj in all_objects:
        registry.register(obj)


def preload_registered_features():
    """
    Load phases and commands from all registered features.
    """
    load_bootstrap_config()
    all_features = features.all()
    enabled_features = [f for f in all_features if not f.disabled]  # type: Iterable[Feature]
    register_objects(enabled_features, lambda f: f.phases, phases)
    register_objects(enabled_features, lambda f: f.commands, commands)
    return enabled_features


def load_registered_features(preload=True):
    """
    Load all registered features.
    """
    if preload:
        load_bootstrap_config()

    all_features = features.all()

    for feature in all_features:
        feature.before_load()

    try:
        for feature in all_features:
            feature.configure()
    except ConfigureSecondPassException:
        config.clear()
        load_bootstrap_config()

        for feature in all_features:
            feature.configure()

    # migrations.compat(config.data)

    enabled_features = [f for f in all_features if not f.disabled]  # type: Iterable[Feature]

    if preload:
        register_objects(enabled_features, lambda f: f.phases, phases)
        register_objects(enabled_features, lambda f: f.commands, commands)

    register_objects(enabled_features, lambda f: [a for a in f.actions if not a.disabled], actions)
    register_objects(enabled_features, lambda f: f.binaries, binaries)
    register_objects(enabled_features, lambda f: f.services, services)

    for feature in enabled_features:
        feature.after_load()

    return enabled_features


def configure_logging(level: Union[str, int] = logging.INFO):
    """
    Configure context logger.
    """
    verboselogs.install()

    log_colors = dict(default_log_colors)
    log_colors['NOTICE'] = 'thin_white'
    log_colors['VERBOSE'] = 'thin_white'
    log_colors['DEBUG'] = 'thin_cyan'
    log_colors['SUCCESS'] = 'bold_green'
    log_colors['SPAM'] = 'bold_red'

    class CustomFormatter(ColoredFormatter):
        """
        Custom context logger formatter.
        """

        def format(self, record):
            record.simplename = record.name.rsplit(".", 1)[-1]
            return super().format(record)

    handler = logging.StreamHandler()
    handler.setFormatter(CustomFormatter(
        '%(log_color)s[%(simplename)s] %(message)s',
        log_colors=log_colors))

    logger = logging.getLogger("ddb")
    logger.handlers.clear()
    logger.setLevel(level)
    logger.addHandler(handler)


def handle_watch():
    """
    Handle watch option
    """
    watch_supports = []  # type:List[Union[Action, WatchSupport]]
    for action in actions.all():
        if isinstance(action, WatchSupport):
            watch_supports.append(action)

    if watch_supports:
        for watch_support in watch_supports:
            logging.getLogger("ddb.watch").info("Initializing %s watcher ...", watch_support.name)
            watch_support.start_watching()
        _watch_started_event.set()
        try:
            context.watching = True
            logging.getLogger("ddb.watch").warning("Watching ... (CTRL+C to terminate)")
            while not _watch_stop_event.wait(1):
                pass
        except KeyboardInterrupt:
            pass
        finally:
            context.watching = False
        logging.getLogger("ddb.watch").warning("Terminating ...")
        for watch_support in watch_supports:
            watch_support.stop_watching()

        for watch_support in watch_supports:
            watch_support.join_watching()
    else:
        logging.getLogger("ddb.watch").warning("Watching is supported by none enabled features")


class ParseCommandLineException(Exception):
    """
    Exception raised when an invalid command line is given.
    """

    def __init__(self, opts: ArgumentParser, parsed_args: Namespace, unknown_args: List[str]):
        super().__init__()
        self.opts = opts
        self.parsed_args = parsed_args
        self.unknown_args = unknown_args


def parse_command_line(args: Optional[Sequence[str]] = None):
    """
    Parse command line arguments, returning the command to execute, args and unknown args.
    """
    opts = ArgumentParser()
    opts.add_argument('-v', '--verbose', action="store_true",
                      default=config.data.get('defaults.verbose', False),
                      help="Enable more logs")
    opts.add_argument('-vv', '--very-verbose', action="store_true",
                      default=config.data.get('defaults.very_verbose', False),
                      help="Enable even more logs")
    opts.add_argument('-s', '--silent', action="store_true",
                      default=config.data.get('defaults.silent', False),
                      help="Disable all logs")
    opts.add_argument('-x', '--exceptions', action="store_true",
                      default=config.data.get('defaults.exceptions', False),
                      help="Display exceptions on errors")
    opts.add_argument("-c", "--clear-cache", action="store_true",
                      default=config.data.get('defaults.clear_cache', False),
                      help="Clear all used caches")
    opts.add_argument('-w', '--watch', action="store_true",
                      default=config.data.get('defaults.watch', False),
                      help="Enable watch mode (hot reload of generated files)")
    opts.add_argument('-ff', '--fail-fast', action="store_true",
                      default=config.data.get('defaults.fail_fast', False),
                      help="Stop on first error")
    opts.add_argument('--version', action="store_true", help='Display the ddb version and check for new ones.')

    command_parsers = {}

    subparsers = opts.add_subparsers(dest="command", help='Available commands')
    for command in commands.all():
        parser = command.add_parser(subparsers)
        command.configure_parser(parser)
        command_parsers[command.name] = parser

    parsed_args, unknown_args = opts.parse_known_args(args)

    log_level = 'INFO'
    if parsed_args.very_verbose:
        log_level = 'DEBUG'
    elif parsed_args.verbose:
        log_level = 'VERBOSE'
    elif parsed_args.silent:
        log_level = 'CRITICAL'

    configure_logging(log_level)

    if not parsed_args.command:
        raise ParseCommandLineException(opts, parsed_args, unknown_args)

    command = commands.get(parsed_args.command)

    if unknown_args and not command.allow_unknown_args:
        msg = _('unrecognized arguments: %s')
        opts.error(msg % ' '.join(unknown_args))

    return command, parsed_args, unknown_args


def handle_command_line(command: Command):
    """
    Execute the command and handle additional given arguments like watch mode
    """
    execute_command(command)

    if config.args.watch:
        handle_watch()


def _register_action_in_event_bus(action: Action, binding: Union[Callable, str, EventBinding], fail_fast=False):
    """
    Register a single event binding. It supports name property added by @event decorator on events callable.
    """
    if callable(binding) and hasattr(binding, "name"):
        binding = binding.name
    if isinstance(binding, str):
        binding = EventBinding(binding)
    if callable(binding.event) and hasattr(binding.event, "name"):
        binding.event = binding.event.name

    bus.on(binding.event, action_event_binding_runner_factory(action,
                                                              binding.event,
                                                              to_call=binding.call,
                                                              event_processor=binding.processor,
                                                              fail_fast=fail_fast).run)


def register_actions_in_event_bus(fail_fast=False):
    """
    Register actions into event bus.
    """
    sorted_actions = sorted(actions.all(), key=lambda x: x.order)

    for action in sorted_actions:
        if isinstance(action.event_bindings, (str, EventBinding)) or callable(action.event_bindings):
            _register_action_in_event_bus(action, action.event_bindings, fail_fast)
        else:
            for event_binding in action.event_bindings:
                _register_action_in_event_bus(action, event_binding, fail_fast)


def main(args: Optional[Sequence[str]] = None,  # pylint:disable=too-many-statements
         reset_disabled=False,
         before_handle_command_line=None):
    """
    Load all features and handle command line
    """
    global _watch_started_event, _watch_stop_event  # pylint:disable=global-statement
    _watch_started_event = threading.Event()
    _watch_stop_event = threading.Event()
    initial_cwd = os.getcwd()
    try:
        load_plugins()
        bootstrap_register_features()
        preload_registered_features()

        try:
            command, args, unknown_args = parse_command_line(args)
            config.args = args
            config.unknown_args = unknown_args
        except ParseCommandLineException as exc:
            config.args = exc.parsed_args
            config.unknown_args = exc.unknown_args
            if not config.args.version:
                exc.opts.print_help()
                raise

        load_registered_features(False)
        register_actions_in_event_bus(config.args.fail_fast)

        if config.args.version:
            events.main.version(silent=config.args.silent)
            return []

        prepare_project_home()
        register_default_caches()

        def on_config_reloaded():
            global _watch_started_event, _watch_stop_event  # pylint:disable=global-statement
            _watch_stop_event.set()
            _watch_started_event = threading.Event()
            _watch_stop_event = threading.Event()

            for action in actions.all():
                if hasattr(action, "destroy") and callable(action.destroy):
                    action.destroy()
                if hasattr(action, "initialized") and action.initialized:
                    action.initialized = False

            context.reset()

            handle_command_line(command)

        bus.on(events.config.reloaded.name, on_config_reloaded)  # pylint:disable=no-member

        events.main.start(command=command)
        handle_command_line(command)

        events.main.terminate(command=command)

        return context.exceptions
    except ExpectedError as exception:
        return [exception]
    except Exception as exception:  # pylint:disable=broad-except
        try:
            clear_caches()
        except Exception as clear_cache_exception:  # pylint:disable=broad-except
            logging.getLogger('ddb').exception("An error has occured while clearing caches",
                                               exc_info=clear_cache_exception)
        raise exception
    finally:
        os.chdir(initial_cwd)
        if not reset_disabled:
            reset()


def wait_watch_started(timeout: Optional[float] = None):
    """
    Wait for watch mode to be started.
    """
    return _watch_started_event.wait(timeout)


def stop_watch():
    """
    Stop watch mode.
    """
    _watch_stop_event.set()


def clear_caches():
    """
    Clear all caches
    """
    for cache in caches.all():
        cache.clear()
        cache.flush()


def reset():
    """
    Close all caches and reset registries to run main method again
    """
    bus.clear()

    features.close()
    phases.close()
    commands.close()
    actions.close()
    binaries.close()
    services.close()
    caches.close()

    context.reset()
    config.reset()

    reset_available_features()


def console_script():  # pragma: no cover
    """
    Console script entrypoint
    """
    try:
        exceptions = main()
    except ParseCommandLineException:
        sys.exit(1)
    if exceptions:
        sys.exit(1)


if __name__ == '__main__':  # pragma: no cover
    console_script()



# -*- coding: utf-8 -*-
import os

import yaml
from ddb.config import Config
from ddb.config.flatten import flatten


def test_defaults():
    config = Config()
    assert config.env_prefix == 'DDB'
    assert config.filenames == ('ddb', 'ddb.local')
    assert config.extensions == ('yml', 'yaml')
    assert len(config.paths) == 3


def test_load_and_clear(data_dir):
    Config.defaults = None
    Config.overrides = None

    ddb_home, home, project = os.path.join(data_dir, 'load', 'ddb_home'), \
                              os.path.join(data_dir, 'load', 'home'), \
                              os.path.join(data_dir, 'load', 'project')

    with open(os.path.join(data_dir, 'load', 'expect.yml'), 'rb') as stream:
        expected = yaml.load(stream, Loader=yaml.FullLoader)

    config = Config(paths=(ddb_home, home, project))
    config.load()

    assert config.data == expected

    config.clear()
    assert not config.data


def test_load_with_env_overrides(data_dir):
    Config.defaults = None
    Config.overrides = None

    ddb_home, home, project = os.path.join(data_dir, 'load', 'ddb_home'), \
                              os.path.join(data_dir, 'load', 'home'), \
                              os.path.join(data_dir, 'load', 'project')

    os.environ['DDB_OVERRIDE_SOME_DEEP'] = "env"
    os.environ['DDB_OVERRIDE_SOME_LIST[1]'] = "replaced"

    with open(os.path.join(data_dir, 'load', 'expect_env.yml'), 'rb') as stream:
        expected = yaml.load(stream, Loader=yaml.FullLoader)

    config = Config(paths=(ddb_home, home, project))
    config.load()

    assert config.data == expected


def test_load_env_variables(data_dir):
    env = os.path.join(data_dir, 'load', 'env')

    config = Config(paths=(env,))
    config.load()

    assert os.environ.get('FOO') == 'bar'
    assert 'env' not in config.data
    assert 'another' in config.data


def test_flatten(data_dir):
    config = Config()
    config.data = {'empty_tags': [], 'tags': ['test'],
                   'app': {'another': 'value',
                           'some': {'empty_tags2': [], 'tags2': ['test2.a', 2],
                                    'complex_list': ['simple', {'key': 'value'}]}}}

    ret = flatten(config.data)
    assert ret == {'tags[0]': 'test', 'app.another': 'value', 'app.some.tags2[0]': 'test2.a',
                   'app.some.tags2[1]': 2, 'app.some.complex_list[0]': 'simple',
                   'app.some.complex_list[1].key': 'value'}

    ret = flatten(config.data, keep_primitive_list=True)
    assert ret == {'empty_tags': [], 'tags': ['test'], 'app.another': 'value', 'app.some.empty_tags2': [],
                   'app.some.tags2': ['test2.a', 2],
                   'app.some.complex_list[0]': 'simple', 'app.some.complex_list[1].key': 'value'}



import os
import zipfile

import yaml
from dotty_dict import Dotty
from pytest_mock import MockerFixture

from ddb.__main__ import main
from ddb.config import Config
from ddb.feature.version import is_git_repository


class TestDockerImageTag:
    def test_image_tag_from_git_tag_jsonnet(self, project_loader, mocker: MockerFixture):
        Config.defaults = None

        mocker.patch('ddb.feature.version.is_git_repository', is_git_repository)

        project_loader("image_tag_from_git_tag")

        if os.path.exists("repo.zip"):
            with zipfile.ZipFile("repo.zip", 'r') as zip_ref:
                zip_ref.extractall(".")

        main(["configure"])

        assert os.path.exists("docker-compose.yml")

        with open("docker-compose.yml") as f:
            docker_compose = yaml.load(f, yaml.SafeLoader)
            assert Dotty(docker_compose).get('services.node.image') == 'some-registry/node:some-tag'

    def test_image_tag_from_git_branch_jsonnet(self, project_loader, mocker: MockerFixture):
        Config.defaults = None

        mocker.patch('ddb.feature.version.is_git_repository', is_git_repository)

        project_loader("image_tag_from_git_branch")

        if os.path.exists("repo.zip"):
            with zipfile.ZipFile("repo.zip", 'r') as zip_ref:
                zip_ref.extractall(".")

        main(["configure"])

        with open("docker-compose.yml") as f:
            docker_compose = yaml.load(f, yaml.SafeLoader)
            assert Dotty(docker_compose).get('services.node.image') == 'some-registry/node:some-branch'

    def test_image_tag_from_git_disabled_jsonnet(self, project_loader, mocker: MockerFixture):
        Config.defaults = None

        mocker.patch('ddb.feature.version.is_git_repository', is_git_repository)

        project_loader("image_tag_from_git_disabled")

        if os.path.exists("repo.zip"):
            with zipfile.ZipFile("repo.zip", 'r') as zip_ref:
                zip_ref.extractall(".")

        main(["configure"])

        with open("docker-compose.yml") as f:
            docker_compose = yaml.load(f, yaml.SafeLoader)
            assert Dotty(docker_compose).get('services.node.image') == 'some-registry/node'



# -*- coding: utf-8 -*-
import os
import re
import shutil
import sys
from datetime import date
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Optional
from urllib.error import HTTPError

import requests
import yaml
from dotty_dict import Dotty
from progress.bar import IncrementalBar
from semver import VersionInfo

from ddb import __version__
from ddb.action import Action, InitializableAction
from ddb.cache import caches, register_global_cache
from ddb.config import config
from ddb.event import events
from ddb.utils.file import force_remove
from ddb.utils.table_display import get_table_display
from .. import features
from ...action.action import EventBinding
from ...action.runner import FailFastError, ExpectedError
from ...command import Command
from ...config.flatten import flatten
from ...context import context


def get_latest_release(github_repository: str):
    """
    Retrieve latest release version and tag_name from GitHub API
    :param github_repository github repository to check
    :return: Version from tag_name retrieved from GitHub API
    """
    response = requests.get('https://api.github.com/repos/{}/releases/latest'.format(github_repository))
    try:
        response.raise_for_status()
        tag_name = response.json().get('tag_name')
        version = tag_name
        if version and version.startswith('v'):
            version = version[1:]
        return version, tag_name
    except HTTPError:  # pylint:disable=bare-except
        return None, None


def get_current_version():
    """
    Get the current version
    :return:
    """
    return __version__


def print_version(github_repository, silent=False):
    """
    Print the version and informations.
    :return:
    """
    if silent:
        print(get_current_version())
        return

    blocks = []
    header_block = ['ddb ' + get_current_version()]
    blocks.append(header_block)

    last_release, tag_name = get_latest_release(github_repository)

    if last_release and get_current_version() < last_release:
        blocks.append(_build_update_header(last_release))
        blocks.append(_build_update_details(github_repository, tag_name))
    blocks.append([
        'Please report any bug or feature request at',
        'https://github.com/inetum-orleans/docker-devbox-ddb/issues'
    ])
    print(get_table_display(blocks))


_version_re = re.compile(
    r"""[vV]?
        (?P<major>0|[1-9]\d*)
        (\.
        (?P<minor>0|[1-9]\d*)
        (\.
            (?P<patch>0|[1-9]\d*)
        )?
        )?
    """,
    re.VERBOSE,
)


def coerce_version(version):
    """
    Convert an incomplete version string into a semver-compatible VersionInfo
    object

    * Tries to detect a "basic" version string (``major.minor.patch``).
    * If not enough components can be found, missing components are
        set to zero to obtain a valid semver version.

    :param str version: the version string to convert
    :return: a tuple with a :class:`VersionInfo` instance (or ``None``
        if it's not a version) and the rest of the string which doesn't
        belong to a basic version.
    :rtype: tuple(:class:`VersionInfo` | None, str)
    """
    match = _version_re.search(version)
    if not match:
        return None, version

    ver = {
        key: 0 if value is None else value for key, value in match.groupdict().items()
    }

    ver = VersionInfo(**ver)
    rest = match.string[match.end():]
    return ver, rest


def is_version_greater(reference, version):
    """
    Check if version is greater than reference.
    :param reference:
    :param version:
    :return:
    """
    return coerce_version(reference) < coerce_version(version)


def is_version_greater_or_equal(reference, version):
    """
    Check if version is greater or equal than reference.
    :param reference:
    :param version:
    :return:
    """
    return coerce_version(reference) < coerce_version(version)


def check_for_update(github_repository: str, output=False, details=False):
    """
    Check if a new version is available on github.
    :param github_repository github repository to check
    :param output: if True, new version information will be displayed.
    :param details: if True, will display more details.
    :return: Version of the latest release if it doesn't match the current one.
    """
    last_release, tag_name = get_latest_release(github_repository)

    if last_release and is_version_greater(get_current_version(), last_release):
        if output:
            blocks = [_build_update_header(last_release)]
            if details:
                row = _build_update_details(github_repository, tag_name)
                blocks.append(row)
                print(get_table_display(blocks))
            else:
                for block in blocks:
                    for row in block:
                        print(row)
        return last_release, tag_name
    return None, None


def _build_update_header(version):
    return ['A new version is available: {}'.format(version)]


def _build_update_details(github_repository, tag_name):
    row = []
    update_tip = _build_update_tip()
    if update_tip:
        row.append(update_tip)
    row.extend((
        'For more information, check the following links:',
        'https://github.com/{}/releases/tag/{}'.format(github_repository, tag_name),
        'https://github.com/{}/releases/tag/{}/CHANGELOG.md'.format(github_repository, tag_name),
    ))
    return row


def _build_update_tip():
    if is_binary():
        return 'run "ddb self-update" command to update.'
    return ''


def is_binary():
    """
    Check if current process is binary.
    :return:
    """
    return getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS')


def get_local_binary_path():
    """
    Get the binary path
    :return:
    """
    context.log.debug("sys.argv[0]: %s", sys.argv[0])
    context.log.debug("config.cwd: %s", config.cwd)
    context.log.debug("sys.frozen: %s", getattr(sys, 'frozen', False))
    context.log.debug("sys._MEIPASS: %s", getattr(sys, '_MEIPASS', ''))
    context.log.debug("sys.executable: %s", sys.executable)
    context.log.debug("__file__: %s", __file__)

    local_binary_path = None
    if sys.executable and os.path.isfile(sys.executable):
        context.log.debug("sys.executable match an existing file. Using this value as binary path.")
        local_binary_path = sys.executable

    if not local_binary_path:
        if os.path.isabs(sys.argv[0]):
            context.log.debug("sys.argv[0] is absolute. Using this value as binary path.")
            local_binary_path = sys.argv[0]

    if not local_binary_path:
        argv0_basename = os.path.basename(sys.argv[0])
        if argv0_basename == sys.argv[0]:
            context.log.debug("sys.argv[0] is basename only. Use configuration values to find expected binary path.")
            local_binary_path = os.path.join(config.paths.home, "bin", sys.argv[0])

    if not local_binary_path:
        if config.cwd:
            context.log.debug("config.cwd is defined. Use configuration values to find expected binary path.")
            local_binary_path = os.path.join(config.cwd, sys.argv[0])

    context.log.debug("local_binary_path: %s", local_binary_path)

    if not local_binary_path:
        raise ValueError("Can't find local binary path to update.")
    return local_binary_path


def get_binary_destination_path(binary_path: str):
    """
    Get binary path destination
    :param binary_path:
    :return:
    """
    if binary_path.endswith('.py') \
            and Path(binary_path).read_bytes().startswith(b'#!/usr/bin/env python'):
        # Avoid removing main source file when running on development.
        binary_path = binary_path[:-3] + ".bin"

    for qualifier in ['-linux', '-alpine', '-macos', '-windows']:
        binary_path = binary_path.replace(qualifier, '')
    return binary_path


class FeaturesAction(Action):
    """
    Display all features
    """

    @property
    def event_bindings(self):
        return events.phase.features

    @property
    def name(self) -> str:
        return "core:features"

    @staticmethod
    def execute():
        """
        Execute action
        """
        enabled_features = [f for f in features.all() if not f.disabled]

        for feature in enabled_features:
            print("%s: %s" % (feature.name, feature.description))


class ConfigAction(Action):
    """
    Dump configuration
    """

    @property
    def event_bindings(self):
        return events.phase.config

    @property
    def name(self) -> str:
        return "core:config"

    @staticmethod
    def execute():
        """
        Execute action
        """
        configuration, configuration_files = ConfigAction._get_configuration_files(
            config.args.full, config.args.files
        )

        if config.args.property:
            configuration, configuration_files = ConfigAction._handle_property(configuration, configuration_files)

        if config.args.value:
            return ConfigAction._print_config_value(configuration, config.args.property)

        if config.args.variables:
            return ConfigAction._print_config_variables(configuration, configuration_files)

        return ConfigAction._print_config_yaml(configuration, configuration_files)

    @staticmethod
    def _handle_property(configuration, configuration_files):
        configuration, configuration_files = ConfigAction._get_configurations_for_prop(
            config.args.property,
            configuration,
            configuration_files
        )
        if configuration is None and not config.args.full:
            configuration, configuration_files = ConfigAction._get_configuration_files(
                True, config.args.files
            )

            configuration, configuration_files = ConfigAction._get_configurations_for_prop(
                config.args.property,
                configuration,
                configuration_files
            )
        if configuration is not None:
            root_config = Dotty({})
            root_config[config.args.property] = configuration
            configuration = dict(root_config)
        if configuration_files:
            root_configuration_files = {}
            for k, configuration_file in configuration_files.items():
                root_configuration_file = Dotty({})
                root_configuration_file[config.args.property] = configuration_file
                root_configuration_files[k] = dict(root_configuration_file)
            configuration_files = root_configuration_files
        return configuration, configuration_files

    @staticmethod
    def _print_config_yaml(configuration, configuration_files):
        if config.args.files and configuration_files:
            for file, configuration_file in configuration_files.items():
                print(f"--- # {file}")
                print(yaml.safe_dump(configuration_file))
        else:
            if isinstance(configuration, (dict, list)):
                print(yaml.safe_dump(configuration))
            elif configuration is not None:
                print(configuration)

    @staticmethod
    def _print_config_variables(configuration, configuration_files):
        if config.args.files and configuration_files:
            for index, (file, configuration_file) in enumerate(configuration_files.items()):
                if index > 0:
                    print()
                print(f"# {file}")
                flat = flatten(Dotty(configuration_file), keep_primitive_list=True)
                for key in sorted(flat.keys()):
                    print(f"{key}: {flat[key]}")
        else:
            flat = flatten(Dotty(configuration), keep_primitive_list=True)
            for key in sorted(flat.keys()):
                print(f"{key}: {flat[key]}")

    @staticmethod
    def _print_config_value(configuration, prop):
        if configuration is None:
            raise ValueError(f"{prop} not found in configuration.")
        dotty_configuration = Dotty(configuration)
        if prop and prop not in dotty_configuration:
            raise ValueError(f"{prop} not found in configuration.")
        value = dotty_configuration[prop] if prop and prop in dotty_configuration else configuration
        print(value)

    @staticmethod
    def _get_configurations_for_prop(prop, configuration, configuration_files):
        prop_configuration = Dotty(configuration).get(prop)

        if configuration_files:
            prop_configuration_files = {}
            for file, configuration_file in configuration_files.items():
                prop_configuration_file = Dotty(configuration_file).get(prop)
                if prop_configuration_file:
                    prop_configuration_files[file] = prop_configuration_file
        else:
            prop_configuration_files = configuration_files

        return prop_configuration, prop_configuration_files

    @staticmethod
    def _prune_default_configuration(default_configuration, flat_file_configuration):
        for file_key in flat_file_configuration.values():
            if file_key in default_configuration:
                default_configuration.pop(file_key, None)
                parent_k = file_key
                while '.' in parent_k:
                    (parent_k, _) = parent_k.rsplit('.', 1)
                    parent_value = default_configuration.get(parent_k)
                    if parent_value is None or not parent_value:
                        default_configuration.pop(parent_k, None)
                    else:
                        break

    @staticmethod
    def _get_configuration_files(full, files):
        default_configuration = dict(config.data.copy())

        if full and not files:
            configuration_files = None
            return default_configuration, configuration_files

        configuration, configuration_files = config.read()
        if full:
            for file_configuration in configuration_files.values():
                flat_file_configuration = flatten(file_configuration)
                ConfigAction._prune_default_configuration(default_configuration, flat_file_configuration)

            tmp = {'default': default_configuration}
            tmp.update(configuration_files)
            configuration_files = tmp
        return configuration, configuration_files


class EjectAction(Action):
    """
    Dump configuration
    """

    @property
    def event_bindings(self):
        def eject_processor(source: Optional[str], target: str):
            if not source or not \
                    os.path.abspath(source).startswith(config.paths.project_home):
                return False
            return (), {"source": source, "target": target}

        return (
            EventBinding(events.file.generated, self.delete_generated_source, eject_processor)
        )

    @property
    def name(self) -> str:
        return "core:eject"

    @property
    def order(self):
        return 1024 * 1024

    @property
    def disabled(self) -> bool:
        if not config.eject:
            return True
        return super().disabled

    @staticmethod
    def delete_generated_source(source: Optional[str], target: str):
        """
        Execute action
        """
        force_remove(source)
        events.file.deleted(source)
        events.file.deleted(target)


class ReloadConfigAction(Action):
    """
    In watch mode, reload config if one configuration file is changed.
    """

    @property
    def event_bindings(self):
        def config_file_processor(file: str):
            if os.path.abspath(file) in config.files:
                return (), {"file": file}
            return None

        return (
            EventBinding(events.file.found, self.execute, config_file_processor)
        )

    @property
    def disabled(self) -> bool:
        return 'watch' not in config.args or not config.args.watch

    @property
    def name(self) -> str:
        return "core:reload-config"

    @staticmethod
    def execute(file: str):
        """
        Execute action
        """
        if context.watching:
            data, _ = config.read()
            try:
                context.log.info("Configuration file has changed.")
                config.clear()
                config.load_from_data(data)
                all_features = features.all()
                for feature in all_features:
                    feature.configure()
                context.log.info("Configuration has been reloaded.")
                events.config.reloaded()
            except Exception as exc:  # pylint:disable=broad-except
                context.log.warning("Configuration has fail to reload: %s", str(exc))
                return


class VersionAction(Action):
    """
    Display display version information when --version flag is used.
    """

    @property
    def name(self) -> str:
        return "core:version"

    @property
    def event_bindings(self):
        return events.main.version

    @staticmethod
    def execute(silent: bool):
        """
        Check for updates
        :param command command name
        :return:
        """
        github_repository = config.data.get('core.github_repository')
        print_version(github_repository, silent)


class CheckForUpdateAction(InitializableAction):
    """
    Check if a new version is available on github.
    """

    def initialize(self):
        register_global_cache('core.check_for_update.version')

    @property
    def name(self) -> str:
        return "core:check-for-update"

    @property
    def event_bindings(self):
        return events.main.terminate

    @staticmethod
    def execute(command: Command):
        """
        Check for updates
        :param command command name
        :return:
        """
        if not command.avoid_stdout and config.data.get('core.check_updates'):
            cache = caches.get('core.check_for_update.version')
            last_check = cache.get('last_check', None)
            today = date.today()

            if last_check is None or last_check < today:
                github_repository = config.data.get('core.github_repository')
                check_for_update(github_repository, True, True)

            cache.set('last_check', today)


class RequiredVersionError(FailFastError, ExpectedError):
    """
    Exception that should be raised when the current version doesn't fullfil the required one.
    """

    def log_error(self):
        context.log.error(str(self))


class CheckRequiredVersion(Action):
    """
    Check if a new version is available on github.
    """

    @property
    def name(self) -> str:
        return "core:check-required-version"

    @property
    def event_bindings(self):
        return events.main.start

    @staticmethod
    def execute(command: Command):
        """
        Check for updates
        :param command command name
        :return:
        """
        if command.name not in ['self-update']:
            required_version = config.data.get('core.required_version')
            if not required_version:
                return
            if is_version_greater_or_equal(get_current_version(), required_version):
                update_tip = _build_update_tip()
                if update_tip:
                    update_tip = ' ' + update_tip
                raise RequiredVersionError(
                    "This project requires ddb {}+. Current version is {}.{}".format(required_version,
                                                                                     get_current_version(),
                                                                                     update_tip))


class SelfUpdateAction(Action):
    """
    Self update ddb if a newer version is available.
    """

    @property
    def name(self) -> str:
        return "core:selfupdate"

    @property
    def event_bindings(self):
        return events.phase.selfupdate

    def execute(self):
        """
        Execute action
        """
        github_repository = config.data.get('core.github_repository')

        if not is_binary():
            print('ddb is running from a package mode than doesn\'t support self-update.')
            print(
                'You can download binary package supporting it ' +
                'from github: https://github.com/{}/releases'.format(github_repository)
            )
            return

        last_release, tag_name = check_for_update(github_repository, True)
        if not last_release:
            print('ddb is already up to date.')
            if 'force' not in config.args or not config.args.force:
                return

        self.self_update_binary(github_repository, tag_name)

    @staticmethod
    def self_update_binary(github_repository, tag_name):
        """
        Self update the ddb binary
        :param github_repository:
        :param tag_name:
        :return:
        """
        local_binary_path = get_local_binary_path()

        if not os.access(local_binary_path, os.W_OK):
            raise PermissionError(f"You don't have permission to write on ddb binary file. ({local_binary_path})")

        if not os.path.isfile(local_binary_path):
            raise ValueError(f"Local binary path is not a file. ({local_binary_path})")

        release_asset_name = config.data.get('core.release_asset_name')
        if not release_asset_name:
            print('ddb is running from a platform that doesn\'t support binary package mode.')
            return

        url = 'https://github.com/{}/releases/download/{}/{}'.format(github_repository, tag_name, release_asset_name)
        context.log.debug('Downloading ddb asset: %s', url)

        progress_bar = None
        with requests.get(url, stream=True) as response:
            response.raise_for_status()

            with NamedTemporaryFile(delete=False) as tmp:
                if not progress_bar:
                    content_length = int(response.headers['content-length'])
                    progress_bar = IncrementalBar('Downloading', max=content_length, suffix='%(percent)d%%')

                for chunk in response.iter_content(32 * 1024):
                    progress_bar.next(len(chunk))  # pylint:disable=not-callable
                    tmp.write(chunk)
                tmp.flush()

            local_binary_path = get_binary_destination_path(local_binary_path)
            shutil.copymode(local_binary_path, tmp.name)
            force_remove(local_binary_path)  # This is required on windows
            os.rename(tmp.name, local_binary_path)

            progress_bar.finish()

        config.data['core.check_updates'] = False
        print("ddb has been updated.")



# -*- coding: utf-8 -*-
import os
import shutil
from typing import Optional

from wcmatch.glob import globmatch, GLOBSTAR

from ddb.action import Action
from ddb.action.action import EventBinding
from ddb.config import config
from ddb.event import events
from ddb.utils.file import chmod


class PermissionsAction(Action):
    """
    Update file access permissions based on git index
    """

    @property
    def name(self) -> str:
        return "git:permissions"

    @property
    def event_bindings(self):
        def file_found_processor(file: str):
            specs = config.data.get('permissions.specs')
            if specs:
                for spec, mode in specs.items():
                    if globmatch(file, spec, flags=GLOBSTAR):
                        return (), {"file": file, "mode": mode}
            return None

        def file_generated_processor(source: Optional[str], target: str):
            mode = file_found_processor(target)
            if mode is None and source is not None:
                shutil.copymode(source, target)

            return mode

        return (
            EventBinding(events.file.found, chmod, file_found_processor),
            EventBinding(events.file.generated, chmod, file_generated_processor)
        )

    @property
    def disabled(self) -> bool:
        return os.name == 'nt'



import os
import re
from abc import ABC, abstractmethod

from _pytest.capture import CaptureFixture
from ddb.__main__ import main, load_registered_features
from ddb.config import config
from ddb.config.config import ConfigPaths
from ddb.feature import features
from ddb.feature.core import CoreFeature
from ddb.feature.shell import ActivateAction, DeactivateAction, ShellFeature
from ddb.feature.shell.actions import encode_environ_backup
from ddb.feature.shell.integrations import BashShellIntegration, ShellIntegration, CmdShellIntegration
from tests.utilstest import expect_gitignore


class ActivateActionBase(ABC):
    @abstractmethod
    def build_shell_integration(self) -> ShellIntegration:
        pass

    @property
    @abstractmethod
    def export_regex(self):
        pass

    @abstractmethod
    def get_shim_filename(self, shimname: str) -> str:
        pass

    @property
    @abstractmethod
    def unset_regex(self):
        pass

    @property
    @abstractmethod
    def filepath_regex(self):
        pass

    @property
    def additional_environment_variable(self) -> dict:
        return {}

    def test_run(self, capsys: CaptureFixture):
        action = ActivateAction(self.build_shell_integration())
        action.execute()

        capture = capsys.readouterr()
        assert capture.out
        assert not capture.err

        filepath = re.match(self.filepath_regex, capture.out).group(1)
        with open(filepath, 'r', encoding='utf-8') as file:
            script = file.read()

        export_match = re.findall(self.export_regex, script, re.MULTILINE)
        env = dict(export_match)

        assert sorted(env.keys()) == sorted(
            ["DDB_PROJECT_HOME", "DDB_SHELL_ENVIRON_BACKUP"] + list(self.additional_environment_variable.keys()))

        assert "DDB_SHELL_ENVIRON_BACKUP" in env
        assert env["DDB_SHELL_ENVIRON_BACKUP"]

    def test_run_using_config(self, capsys: CaptureFixture):
        config.data["some"] = "testing"
        config.data["another.deep"] = "testing2"

        action = ActivateAction(self.build_shell_integration())
        action.execute()

        capture = capsys.readouterr()
        assert capture.out
        assert not capture.err

        filepath = re.match(self.filepath_regex, capture.out).group(1)
        with open(filepath, 'r', encoding='utf-8') as file:
            script = file.read()

        export_match = re.findall(self.export_regex, script, re.MULTILINE)
        env = dict(export_match)

        assert sorted(env.keys()) == sorted(
            ["DDB_PROJECT_HOME", "DDB_SOME", "DDB_ANOTHER_DEEP",
             "DDB_SHELL_ENVIRON_BACKUP"] + list(self.additional_environment_variable.keys()))

    def test_run_activate_deactivate_project(self, capsys: CaptureFixture, project_loader):
        project_loader("project")

        features.register(CoreFeature())
        features.register(ShellFeature())
        load_registered_features()

        action = ActivateAction(self.build_shell_integration())
        action.execute()

        capture = capsys.readouterr()
        assert capture.out
        assert not capture.err

        filepath = re.match(self.filepath_regex, capture.out).group(1)
        with open(filepath, 'r', encoding='utf-8') as file:
            script = file.read()

        export_match = re.findall(self.export_regex, script, re.MULTILINE)
        env = dict(export_match)

        system_path = env.get('PATH', '')
        first = system_path.split(os.pathsep)[0]

        assert first == os.path.normpath(os.path.join(os.getcwd(), "./bin")) \
               or first == os.path.normpath(os.path.join(os.getcwd(), "./.bin"))

        os.environ.update(env)

        deactivate_action = DeactivateAction(self.build_shell_integration())
        deactivate_action.execute()

        capture = capsys.readouterr()
        assert capture.out
        assert not capture.err

        filepath = re.match(self.filepath_regex, capture.out).group(1)
        with open(filepath, 'r', encoding='utf-8') as file:
            script = file.read()

        export_match = re.findall(self.export_regex, script, re.MULTILINE)
        env = dict(export_match)

        system_path = env.get('PATH', '').split(os.pathsep)
        first = system_path[0]

        assert first != os.path.normpath(os.path.join(os.getcwd(), "./bin")) or \
               first != os.path.normpath(os.path.join(os.getcwd(), "./.bin"))

    def test_run_activate_deactivate_project_prepend_false(self, capsys: CaptureFixture, project_loader):
        project_loader("project_prepend_false")

        features.register(CoreFeature())
        features.register(ShellFeature())
        load_registered_features()

        action = ActivateAction(self.build_shell_integration())
        action.execute()

        capture = capsys.readouterr()
        assert capture.out
        assert not capture.err

        filepath = re.match(self.filepath_regex, capture.out).group(1)
        with open(filepath, 'r', encoding='utf-8') as file:
            script = file.read()

        export_match = re.findall(self.export_regex, script, re.MULTILINE)
        env = dict(export_match)

        system_path = env.get('PATH', '').split(os.pathsep)
        last = system_path[-1]

        assert last == os.path.normpath(os.path.join(os.getcwd(), "./bin")) \
               or last == os.path.normpath(os.path.join(os.getcwd(), "./.bin"))

        os.environ.update(env)

        deactivate_action = DeactivateAction(self.build_shell_integration())
        deactivate_action.execute()

        capture = capsys.readouterr()
        assert capture.out
        assert not capture.err

        filepath = re.match(self.filepath_regex, capture.out).group(1)
        with open(filepath, 'r', encoding='utf-8') as file:
            script = file.read()

        export_match = re.findall(self.export_regex, script, re.MULTILINE)
        env = dict(export_match)

        system_path = env.get('PATH', '').split(os.pathsep)
        last = system_path[-1]

        assert last != os.path.normpath(os.path.join(os.getcwd(), "./bin")) or \
               last != os.path.normpath(os.path.join(os.getcwd(), "./.bin"))

    def test_empty_project_main_aliases(self, project_loader):
        project_loader("aliases")

        main(["configure"])

        alias = "dc.bat" if os.name == "nt" else "dc"
        assert os.path.exists(os.path.join("bin", alias))
        assert not os.path.exists(os.path.join(config.paths.home, "bin", alias))

        with open(os.path.join("bin", alias), 'r') as f:
            content = f.read()
            if os.name != 'nt':
                assert 'docker-compose "$@"' in content
            # TODO: Check the content of the file with an assertion, and test the feature behavior on windows

    def test_empty_project_global_aliases(self, project_loader):
        project_loader("global_aliases")

        main(["configure"])

        alias = "dc.bat" if os.name == 'nt' else "dc"

        assert not os.path.exists(os.path.join(".bin", alias))
        assert os.path.exists(os.path.join(config.paths.home, ".bin", alias))
        assert not expect_gitignore('.gitignore', '../home/.bin/dc')

        with open(os.path.join(config.paths.home, ".bin", alias), 'r') as f:
            content = f.read()
            if os.name != 'nt':
                assert 'docker-compose "$@"' in content
            # TODO: Check the content of the file with an assertion, and test the feature behavior on windows

    def test_global_aliases_projects_in_home(self, project_loader, data_dir):
        def project_in_home_config_provider(root_dir):
            paths = ConfigPaths(ddb_home=os.path.join(root_dir, 'ddb_home'), home=os.path.join(root_dir, 'home'),
                                project_home=os.path.join(root_dir, 'home', 'project'))

            if not [path for path in paths if os.path.isdir(path)]:
                paths = ConfigPaths(ddb_home=None, home=None, project_home=root_dir)

            config.paths = paths
            config.load()

            return config

        project_loader("global_aliases_projects_in_home", config_provider=project_in_home_config_provider)

        main(["configure"])

        alias = "dc.bat" if os.name == 'nt' else "dc"

        assert not os.path.exists(os.path.join(".bin", alias))
        assert os.path.exists(os.path.join(config.paths.home, ".bin", alias))
        assert not expect_gitignore('.gitignore', '../home/.bin/dc')

        with open(os.path.join(config.paths.home, ".bin", alias), 'r') as f:
            content = f.read()
            if os.name != 'nt':
                assert 'docker-compose "$@"' in content
            # TODO: Check the content of the file with an assertion, and test the feature behavior on windows


class TestBashActivateAction(ActivateActionBase):
    export_regex = r"^export (.+?)=[\'\"]?(.+?)[\'\"]?$"
    unset_regex = r"^/unset (.+)()$"
    filepath_regex = r"^(?:.* )?(.+)$"

    def build_shell_integration(self) -> ShellIntegration:
        return BashShellIntegration()

    def get_shim_filename(self, shimname: str) -> str:
        return shimname


class TestCmdActivateAction(ActivateActionBase):
    export_regex = r"^set (.+?)=[\'\"]?(.+?)[\'\"]?$"
    unset_regex = r"^set (.+)=()$"
    filepath_regex = r"^(?:.* )?(.+)$"

    @property
    def additional_environment_variable(self) -> dict:
        return {'NL': '^'}

    def build_shell_integration(self) -> ShellIntegration:
        return CmdShellIntegration()

    def get_shim_filename(self, shimname: str) -> str:
        return "{}.bat".format(shimname)


class DeactivateActionBase(ABC):
    @abstractmethod
    def build_shell_integration(self) -> ShellIntegration:
        pass

    @property
    @abstractmethod
    def export_regex(self):
        pass

    @property
    @abstractmethod
    def unset_regex(self):
        pass

    @property
    @abstractmethod
    def filepath_regex(self):
        pass

    @property
    def additional_environment_variable(self) -> dict:
        return {}

    def test_run(self, capsys: CaptureFixture):
        os.environ['DDB_PROJECT_HOME'] = config.paths.project_home
        os.environ['DDB_SHELL_ENVIRON_BACKUP'] = encode_environ_backup(dict(os.environ))

        action = DeactivateAction(self.build_shell_integration())
        action.execute()

        capture = capsys.readouterr()
        assert capture.out
        assert not capture.err

        filepath = re.match(self.filepath_regex, capture.out).group(1)
        with open(filepath, 'r', encoding='utf-8') as file:
            script = file.read()

        export_match = re.findall(self.export_regex, script, re.MULTILINE)
        exported = dict(export_match)
        assert exported == self.additional_environment_variable

        unset_match = re.findall(self.unset_regex, script, re.MULTILINE)
        unset = dict(unset_match)
        assert len(unset) == 1
        assert 'DDB_SHELL_ENVIRON_BACKUP' in unset.keys()

    def test_run_with_empty_initial_env(self, capsys: CaptureFixture):
        expected_environ = dict()

        os.environ.clear()
        os.environ.update(expected_environ)
        os.environ['DDB_PROJECT_HOME'] = config.paths.project_home
        os.environ['DDB_SHELL_ENVIRON_BACKUP'] = encode_environ_backup(expected_environ)

        action = DeactivateAction(self.build_shell_integration())
        action.execute()

        capture = capsys.readouterr()
        assert capture.out
        assert not capture.err

        filepath = re.match(self.filepath_regex, capture.out).group(1)
        with open(filepath, 'r', encoding='utf-8') as file:
            script = file.read()

        export_match = re.findall(self.export_regex, script, re.MULTILINE)
        exported = dict(export_match)
        assert exported == self.additional_environment_variable

        unset_match = re.findall(self.unset_regex, script, re.MULTILINE)
        unset = dict(unset_match)
        assert unset

        assert sorted(unset.keys()) == sorted(("DDB_PROJECT_HOME", "DDB_SHELL_ENVIRON_BACKUP",))

    def test_run_with_changed_env_variable(self, capsys: CaptureFixture):
        expected_environ = {"DDB_CHANGE": "foo", "DDB_NO_CHANGE": "ok", "DDB_REMOVED": "removed"}

        os.environ.clear()
        os.environ.update(expected_environ)
        os.environ['DDB_PROJECT_HOME'] = config.paths.project_home
        os.environ['DDB_SHELL_ENVIRON_BACKUP'] = encode_environ_backup(expected_environ)

        removed_item = os.environ.pop("DDB_REMOVED")
        os.environ["DDB_CHANGE"] = "bar"
        os.environ["DDB_ADDED"] = "added"

        action = DeactivateAction(self.build_shell_integration())
        action.execute()

        capture = capsys.readouterr()
        assert capture.out
        assert not capture.err

        filepath = re.match(self.filepath_regex, capture.out).group(1)
        with open(filepath, 'r', encoding='utf-8') as file:
            script = file.read()

        export_match = re.findall(self.export_regex, script, re.MULTILINE)
        exported = dict(export_match)
        assert len(exported) == (2 + len(self.additional_environment_variable.keys()))
        assert sorted(exported.keys()) == sorted(
            ["DDB_CHANGE", "DDB_REMOVED"] + list(self.additional_environment_variable.keys()))
        assert sorted(exported.values()) == sorted(
            ["foo", removed_item] + list(self.additional_environment_variable.values()))

        unset_match = re.findall(self.unset_regex, script, re.MULTILINE)
        unset = dict(unset_match)
        assert sorted(unset.keys()) == sorted(("DDB_PROJECT_HOME", "DDB_SHELL_ENVIRON_BACKUP", "DDB_ADDED"))


class TestBashDeactivateAction(DeactivateActionBase):
    export_regex = r"^export (.+?)=[\'\"]?(.+?)[\'\"]?$"
    unset_regex = r"^unset (.+)()$"
    filepath_regex = r"^(?:.* )?(.+)$"

    def build_shell_integration(self) -> ShellIntegration:
        return BashShellIntegration()


class TestCmdDeactivateAction(DeactivateActionBase):
    export_regex = r"^set (.+?)=[\'\"]?(.+?)[\'\"]?$"
    unset_regex = r"^set (.+)=()$"
    filepath_regex = r"^(?:.* )?(.+)$"

    @property
    def additional_environment_variable(self) -> dict:
        return {'NL': '^'}

    def build_shell_integration(self) -> ShellIntegration:
        return CmdShellIntegration()



# -*- coding: utf-8 -*-

from marshmallow import fields, Schema

from ddb.feature.schema import FeatureSchema


class CfsslCertificateRequest(Schema):
    """
    CFSSL certificate request schema
    """
    org_name = fields.String()  # The full legal name of the organization. Do not abbreviate.
    org_unit = fields.String()  # Section of the organization.
    city = fields.String()  # The city where the organization is legally located.
    state = fields.String()  # The state or province where your organization is legally located. Can not be abbreviated.
    country = fields.String()  # The two letter ISO abbreviation for the country.


class CfsslServerSchema(Schema):
    """
    Cfssl server schema
    """
    host = fields.String(required=True, default="localhost")
    port = fields.Integer(required=True, default=7780)
    ssl = fields.Boolean(required=True, default=False)
    verify_cert = fields.Boolean(required=True, default=True)


class CfsslWriterFilenamesSchema(Schema):
    """
    CFSSL writer filenames schema
    """

    private_key = fields.String(required=True, default='%s.key')
    certificate = fields.String(required=True, default='%s.crt')
    certificate_request = fields.String(required=True, default='%s.csr')
    certificate_der = fields.String(required=True, default='%s.crt.der')
    certificate_request_der = fields.String(required=True, default='%s.csr.der')


class CfsslWriterSchema(Schema):
    """
    CFSSL writer schema
    """
    filenames = fields.Nested(CfsslWriterFilenamesSchema(), required=True, default=CfsslWriterFilenamesSchema())


class CfsslSchema(Schema):
    """
    Debug schema
    """
    server = fields.Nested(CfsslServerSchema(), required=True, default=CfsslServerSchema())
    append_ca_certificate = fields.Boolean(required=True, default=True)
    certificate_request = fields.List(fields.Nested(CfsslCertificateRequest()))
    verify_checksum = fields.Boolean(required=True, default=True)
    writer = fields.Nested(CfsslWriterSchema(), required=True, default=CfsslWriterSchema())


class CertsSchema(FeatureSchema):
    """
    Certs feature schema.
    """
    type = fields.String(required=True, default="cfssl")
    destination = fields.String(required=True, default=".certs")
    signer_destinations = fields.List(fields.String(), required=False, default=[])
    cfssl = fields.Nested(CfsslSchema(), required=False, default=CfsslSchema())



# -*- coding: utf-8 -*-
from typing import ClassVar, Iterable

from dotty_dict import Dotty

from ddb.action import Action
from ddb.feature import Feature
from .actions import JinjaAction
from .schema import JinjaSchema
from ...utils.file import TemplateFinder


class JinjaFeature(Feature):
    """
    Render template files with Jinja (https://jinja.palletsprojects.com).
    """

    @property
    def name(self) -> str:
        return "jinja"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core", "file"]

    @property
    def schema(self) -> ClassVar[JinjaSchema]:
        return JinjaSchema

    @property
    def actions(self) -> Iterable[Action]:
        return (
            JinjaAction(),
        )

    def _configure_defaults(self, feature_config: Dotty):
        includes = feature_config.get("includes")
        if includes is None:
            includes = TemplateFinder.build_default_includes_from_suffixes(
                feature_config["suffixes"],
                feature_config["extensions"]
            )
            feature_config["includes"] = includes



import math
from typing import List, Iterable


def _get_content(content: str, line_length: int, centered: bool = True):
    left = ' '
    right = (line_length - len(content) - 1) * ' '

    if centered:
        left = math.floor((line_length - len(content)) / 2) * ' '
        right = math.ceil((line_length - len(content)) / 2) * ' '

    return left + content + right


def _max_length(cells: Iterable[Iterable[str]]) -> int:
    max_length = 0
    for cell in cells:
        for row in cell:
            length = len(row)
            if length > max_length:
                max_length = length
    return max_length


def get_table_display(blocks: Iterable[Iterable[str]], centered: bool = True):
    """
    Convert the input data into a table with centered text
    :param blocks: the blocks of the table. A block is an iterable of rows.
    :param centered: if the cell content is centered or not
    :return:
    """
    line_length = _max_length(blocks) + 2

    content = []
    for block in blocks:
        content.append('+' + (line_length * '-') + '+')
        for row in block:
            content.append('|' + _get_content(row, line_length, centered) + '|')
    if content:
        content.append('+' + (line_length * '-') + '+')
    return '\n'.join(content)



# -*- coding: utf-8 -*-
from typing import Iterable

import pkg_resources
from toposort import toposort_flatten

from . import features
from .certs import CertsFeature
from .cookiecutter import CookiecutterFeature
from .copy import CopyFeature
from .core import CoreFeature
from .docker import DockerFeature
from .feature import Feature
from .file import FileFeature
from .fixuid import FixuidFeature
from .git import GitFeature
from .gitignore import GitignoreFeature
from .jinja import JinjaFeature
from .jsonnet import JsonnetFeature
from .permissions import PermissionsFeature
from .run import RunFeature
from .shell import ShellFeature
from .smartcd import SmartcdFeature
from .symlinks import SymlinksFeature
from .traefik import TraefikFeature
from .version import VersionFeature
from .ytt import YttFeature
from ..config import config
from ..config import migrations

_default_available_features = [CertsFeature(),
                               CopyFeature(),
                               CookiecutterFeature(),
                               CoreFeature(),
                               DockerFeature(),
                               FileFeature(),
                               FixuidFeature(),
                               GitFeature(),
                               GitignoreFeature(),
                               JinjaFeature(),
                               JsonnetFeature(),
                               PermissionsFeature(),
                               RunFeature(),
                               SmartcdFeature(),
                               ShellFeature(),
                               SymlinksFeature(),
                               TraefikFeature(),
                               VersionFeature(),
                               YttFeature()]

_available_features = list(_default_available_features)


def get_sorted_features(available_features: Iterable[Feature] = None):
    """
    Register default features and setuptools entrypoint 'ddb_features' inside features registry.
    Features are registered in order for their dependency to be registered first with a topological sort.
    Withing a command phase, actions are executed in the order of their feature registration.
    """
    if available_features is None:
        available_features = _available_features

    entrypoint_features = {f.name: f for f in available_features}
    for entry_point in pkg_resources.iter_entry_points('ddb_features'):
        feature = entry_point.load()()
        entrypoint_features[feature.name] = feature

    required_dependencies, toposort_data = _prepare_dependencies_data(entrypoint_features)
    _check_missing_dependencies(entrypoint_features, required_dependencies)

    dependencies = config.data.get('dependencies')
    if dependencies:
        for feat, feat_dependencies in dependencies.items():
            if feat not in toposort_data:
                toposort_data[feat] = set()
            for feat_dependency in feat_dependencies:
                toposort_data[feat].add(feat_dependency)

    sorted_feature_names = toposort_flatten(toposort_data, sort=True)
    for feature_name in sorted_feature_names:
        feature = entrypoint_features.get(feature_name)
        if feature:
            yield feature


def _prepare_dependencies_data(entrypoint_features):
    """
    Compute required dependencies and toposort data.
    """
    required_dependencies = {}
    toposort_data = {}
    for name, feat in entrypoint_features.items():
        feat_required_dependencies = []
        dependencies = set()
        for dependency_item in feat.dependencies:
            if not dependency_item.endswith('[optional]'):
                feat_required_dependencies.append(dependency_item)
            else:
                dependency_item = dependency_item[0:len(dependency_item) - len('[optional]')]
            dependencies.add(dependency_item)

        toposort_data[name] = dependencies
        required_dependencies[name] = feat_required_dependencies
    return required_dependencies, toposort_data


def _check_missing_dependencies(entrypoint_features, required_dependencies):
    """
    Check missing required dependencies
    """
    for name, feat_required_dependencies in required_dependencies.items():
        for required_dependency in feat_required_dependencies:
            if required_dependency not in entrypoint_features.keys():
                raise ValueError("A required dependency is missing for " +
                                 name + " feature (" + required_dependency + ")")


def reset_available_features():
    """
    Reset available features to default list.
    """
    global _available_features  # pylint:disable=global-statement
    _available_features = list(_default_available_features)


def append_available_feature(feature):
    """
    Append a feature to available features list.
    """
    global _available_features  # pylint:disable=global-statement,global-variable-not-assigned
    _available_features.append(feature)


def bootstrap_register_features():
    """
    Register all features in order.
    :return:
    """
    config.load()
    features.clear()
    for feature in get_sorted_features():
        features.register(feature)
    config.clear()


def bootstrap_features_configuration():
    """
    Preconfigure features with default bootstrap configuration.
    :return:
    """
    try:
        filenames, extensions = config.filenames, config.extensions

        config.filenames, config.extensions = (), ()

        enabled_features = [f for f in features.all() if not f.disabled]
        for feature in enabled_features:
            feature.configure(bootstrap=True)

    finally:
        config.filenames, config.extensions = filenames, extensions


def load_bootstrap_config():
    """
    Load bootstrap configuration.
    """
    bootstrap_features_configuration()
    config.load(config.data)
    migrations.migrate(config.data)



# -*- coding: utf-8 -*-
from typing import Iterable

from ddb.action import Action
from ddb.feature import Feature
from .actions import CustomAction


class CustomFeature(Feature):
    """
    Custom Feature
    """

    @property
    def name(self) -> str:
        return "custom"

    @property
    def actions(self) -> Iterable[Action]:
        return (
            CustomAction(),
        )



import os
from pathlib import Path

import pytest
from _pytest.capture import CaptureFixture
from pytest_mock import MockerFixture

from ddb import __version__
from ddb.__main__ import main


class TestCore:
    bin = './bin/ddb'

    def test_self_update_no_binary(self, project_loader, capsys: CaptureFixture, mocker: MockerFixture):
        mocker.patch('ddb.feature.core.actions.get_local_binary_path', lambda *args, **kwargs: self.bin)
        mocker.patch('ddb.feature.core.actions.is_binary', lambda *args, **kwargs: False)

        project_loader("empty")

        main(["self-update"])

        assert Path('./bin/ddb').read_bytes() == b"binary"

        outerr = capsys.readouterr()
        assert outerr.err == ""
        assert outerr.out == ('ddb is running from a package mode than doesn\'t support self-update.\n'
                              'You can download binary package supporting it from github: '
                              'https://github.com/inetum-orleans/docker-devbox-ddb/releases\n')

    def test_self_update_up_to_date(self, project_loader, capsys: CaptureFixture, mocker: MockerFixture):
        mocker.patch('ddb.feature.core.actions.get_local_binary_path', lambda *args, **kwargs: self.bin)
        mocker.patch('ddb.feature.core.actions.is_binary', lambda *args, **kwargs: True)
        mocker.patch('ddb.feature.core.actions.get_latest_release',
                     lambda *args, **kwargs: (__version__, 'v' + __version__))

        project_loader("empty")

        main(["self-update"])

        assert Path('./bin/ddb').read_bytes() == b"binary"

        outerr = capsys.readouterr()
        assert outerr.err == ""
        assert outerr.out == 'ddb is already up to date.\n'

    def test_self_update_outdated(self, project_loader, capsys: CaptureFixture, mocker: MockerFixture):
        mocker.patch('ddb.feature.core.actions.get_local_binary_path', lambda *args, **kwargs: self.bin)
        mocker.patch('ddb.feature.core.actions.is_binary', lambda *args, **kwargs: True)
        mocker.patch('ddb.feature.core.actions.get_latest_release', lambda *args, **kwargs: ('1.10.0', 'v1.10.0'))
        mocker.patch('ddb.feature.core.actions.get_current_version', lambda *args, **kwargs: '1.9.2')

        project_loader("empty")

        main(["self-update"])

        assert Path('./bin/ddb').read_bytes() != b"binary"

        outerr = capsys.readouterr()
        assert outerr.err == ""
        assert outerr.out == 'A new version is available: 1.10.0\nddb has been updated.\n'

    @pytest.mark.skip("Should be enabled after 1.10.0 release")
    def test_self_update_up_to_date_force(self, project_loader, capsys: CaptureFixture, mocker: MockerFixture):
        mocker.patch('ddb.feature.core.actions.get_local_binary_path', lambda *args, **kwargs: self.bin)
        mocker.patch('ddb.feature.core.actions.is_binary', lambda *args, **kwargs: True)
        mocker.patch('ddb.feature.core.actions.get_latest_release', lambda *args, **kwargs: ('1.10.0', 'v1.10.0'))
        mocker.patch('ddb.feature.core.actions.get_current_version', lambda *args, **kwargs: '1.9.2')

        project_loader("empty")

        main(["self-update", "--force"])

        assert Path('./bin/ddb').read_bytes() != b"binary"

        outerr = capsys.readouterr()
        assert outerr.err == ""
        assert outerr.out == 'ddb is already up to date.\nddb has been updated.\n'

    def test_self_update_up_to_date(self, project_loader, capsys: CaptureFixture, mocker: MockerFixture):
        mocker.patch('ddb.feature.core.actions.get_local_binary_path', lambda *args, **kwargs: self.bin)
        mocker.patch('ddb.feature.core.actions.is_binary', lambda *args, **kwargs: True)
        mocker.patch('ddb.feature.core.actions.get_latest_release', lambda *args, **kwargs: ('1.3.0', 'v1.3.0'))
        mocker.patch('ddb.feature.core.actions.get_current_version', lambda *args, **kwargs: '1.3.0')

        project_loader("empty")

        main(["self-update"])

        assert Path('./bin/ddb').read_bytes() == b"binary"

        outerr = capsys.readouterr()
        assert outerr.err == ""
        assert outerr.out == 'ddb is already up to date.\n'

    def test_version_up_to_date(self, project_loader, capsys: CaptureFixture, mocker: MockerFixture):
        mocker.patch('ddb.feature.core.actions.get_local_binary_path', lambda *args, **kwargs: self.bin)
        mocker.patch('ddb.feature.core.actions.is_binary', lambda *args, **kwargs: True)
        mocker.patch('ddb.feature.core.actions.get_latest_release', lambda *args, **kwargs: ('1.3.0', 'v1.3.0'))
        mocker.patch('ddb.feature.core.actions.get_current_version', lambda *args, **kwargs: '1.3.0')

        project_loader("empty")

        main(["--version"])

        outerr = capsys.readouterr()
        assert outerr.err == ""
        assert outerr.out == '''+------------------------------------------------------------+
|                         ddb 1.3.0                          |
+------------------------------------------------------------+
|        Please report any bug or feature request at         |
| https://github.com/inetum-orleans/docker-devbox-ddb/issues |
+------------------------------------------------------------+\n'''

    def test_version_outdated_binary(self, project_loader, capsys: CaptureFixture, mocker: MockerFixture):
        mocker.patch('ddb.feature.core.actions.get_local_binary_path', lambda *args, **kwargs: self.bin)
        mocker.patch('ddb.feature.core.actions.is_binary', lambda *args, **kwargs: True)
        mocker.patch('ddb.feature.core.actions.get_latest_release', lambda *args, **kwargs: ('1.3.1', 'v1.3.1'))
        mocker.patch('ddb.feature.core.actions.get_current_version', lambda *args, **kwargs: '1.3.0')

        project_loader("empty")

        main(["--version"])

        outerr = capsys.readouterr()
        assert outerr.err == ""
        assert outerr.out == '''+--------------------------------------------------------------------------------------+
|                                      ddb 1.3.0                                       |
+--------------------------------------------------------------------------------------+
|                          A new version is available: 1.3.1                           |
+--------------------------------------------------------------------------------------+
|                       run "ddb self-update" command to update.                       |
|                   For more information, check the following links:                   |
|       https://github.com/inetum-orleans/docker-devbox-ddb/releases/tag/v1.3.1        |
| https://github.com/inetum-orleans/docker-devbox-ddb/releases/tag/v1.3.1/CHANGELOG.md |
+--------------------------------------------------------------------------------------+
|                     Please report any bug or feature request at                      |
|              https://github.com/inetum-orleans/docker-devbox-ddb/issues              |
+--------------------------------------------------------------------------------------+\n'''

    def test_required_version_eq(self, project_loader, capsys: CaptureFixture, mocker: MockerFixture):
        mocker.patch('ddb.feature.core.actions.get_current_version', lambda *args, **kwargs: '1.3.0')

        project_loader("required-version")

        exceptions = main(["configure"])

        assert not exceptions
        assert os.path.exists('test')

    def test_required_version_lt(self, project_loader, capsys: CaptureFixture, mocker: MockerFixture):
        mocker.patch('ddb.feature.core.actions.get_current_version', lambda *args, **kwargs: '1.2.9')

        project_loader("required-version")

        exceptions = main(["configure"])
        assert len(exceptions) == 1
        assert str(exceptions[0]) == "This project requires ddb 1.3.0+. Current version is 1.2.9."

        outerr = capsys.readouterr()
        assert "This project requires ddb 1.3.0+. Current version is 1.2.9." in outerr.err
        assert outerr.out == ""

        assert not os.path.exists('test')

    def test_required_version_lt_binary(self, project_loader, capsys: CaptureFixture, mocker: MockerFixture):
        mocker.patch('ddb.feature.core.actions.is_binary', lambda *args, **kwargs: True)
        mocker.patch('ddb.feature.core.actions.get_current_version', lambda *args, **kwargs: '1.2.9')

        project_loader("required-version")

        exceptions = main(["configure"])
        assert len(exceptions) == 1
        assert str(exceptions[0]) == "This project requires ddb 1.3.0+. " \
                                     "Current version is 1.2.9. " \
                                     "run \"ddb self-update\" command to update."

        outerr = capsys.readouterr()
        assert "This project requires ddb 1.3.0+. Current version is 1.2.9." in outerr.err
        assert outerr.out == ""

        assert not os.path.exists('test')

    def test_required_version_gt(self, project_loader, capsys: CaptureFixture, mocker: MockerFixture):
        mocker.patch('ddb.feature.core.actions.get_current_version', lambda *args, **kwargs: '1.3.1')

        project_loader("required-version")

        exceptions = main(["configure"])

        assert not exceptions
        assert os.path.exists('test')



# -*- coding: utf-8 -*-
import json
import os
import re
from importlib import import_module
from typing import Tuple, Union, Iterable, Optional

import yaml
from _jsonnet import evaluate_file  # pylint: disable=no-name-in-module

from ddb.action.action import AbstractTemplateAction
from ddb.config import config, migrations
from ddb.config.flatten import flatten
from ddb.feature import features
from ddb.utils.file import TemplateFinder, SingleTemporaryFile


class JsonnetAction(AbstractTemplateAction):
    """
    Render jsonnet files based on filename suffixes.
    """

    @property
    def name(self) -> str:
        return "jsonnet:render"

    def _build_template_finder(self) -> TemplateFinder:
        return TemplateFinder(config.data.get("jsonnet.includes"),
                              config.data.get("jsonnet.excludes"),
                              config.data.get("jsonnet.suffixes"))

    def _render_template(self, template: str, target: str) -> Iterable[Tuple[Union[str, bytes, bool], str]]:
        """
        Render jsonnet template
        """
        evaluated = self._evaluate_jsonnet(template)

        multiple_file_output, multiple_file_dir = JsonnetAction._parse_multiple_header(template, target)
        if multiple_file_output:
            for (filename, content) in json.loads(evaluated).items():
                evaluated_target_path = os.path.join(multiple_file_dir, filename)
                evaluated_target_parent_path = os.path.dirname(evaluated_target_path)
                if evaluated_target_parent_path and not os.path.exists(evaluated_target_parent_path):
                    os.makedirs(evaluated_target_parent_path)

                yield content, evaluated_target_path
        else:
            ext = os.path.splitext(target)[-1]
            if ext.lower() in ['.yaml', '.yml']:
                data = json.loads(evaluated)
                self._do_postprocess(data)
                evaluated = yaml.safe_dump(data)
            else:
                if '__post_processors__' in evaluated:
                    data = json.loads(evaluated)
                    if self._do_postprocess(data):
                        evaluated = json.dumps(data)

            yield evaluated, target

    @staticmethod
    def _do_postprocess(data):
        """
        Post process jsonnet output with python functions.
        """
        if isinstance(data, dict) and '__post_processors__' in data.keys():
            for post_processor_str in data.pop('__post_processors__'):
                if '__value__' in data:
                    data = data.pop('__value__')
                module_name, func_name = post_processor_str.rsplit('.', 1)
                if module_name == 'ddb.feature.jsonnet.docker':
                    mod = import_module(module_name)
                    post_processor = getattr(mod, func_name)
                    post_processor(data)
                    return True
        return False

    def _autofix_render_error(self,
                              template: str,
                              target: str,
                              original_template: str,
                              render_error: Exception) -> Optional[str]:
        property_name_match = re.match('RUNTIME ERROR: undefined external variable: (.*)\n', str(render_error))
        if property_name_match:
            property_name = property_name_match.group(1)

            property_migration = migrations.get_migration_from_old_config_key(property_name)

            if property_migration and not property_migration.requires_value_migration:
                property_migration.warn(original_template)

                with open(template, "r", encoding="utf-8") as template_file:
                    initial_template_data = template_file.read()
                    template_data = initial_template_data
                    template_data = template_data.replace('std.extVar("' + property_migration.old_config_key + '")',
                                                          'std.extVar("' + property_migration.new_config_key + '")')
                    template_data = template_data.replace("std.extVar('" + property_migration.old_config_key + "')",
                                                          "std.extVar('" + property_migration.new_config_key + "')")

                if initial_template_data != template_data:
                    with SingleTemporaryFile("ddb", "migration", "jsonnet",
                                             mode="w",
                                             prefix="",
                                             suffix="." + os.path.basename(original_template),
                                             encoding="utf-8") as tmp_file:
                        tmp_file.write(template_data)
                        return tmp_file.name

        return None

    @staticmethod
    def _evaluate_jsonnet(template_path):
        data = config.data.copy()

        vars_config = flatten(data, stop_for_features=features.all())
        codes_config = flatten(data, keep_primitive_list=True, stop_for_features=features.all())
        codes_config['_config.eject'] = config.eject
        codes_config['_config.args'] = vars(config.args)
        codes_config['_config.unknown_args'] = config.unknown_args
        ext_vars = {k: v for (k, v) in vars_config.items() if isinstance(v, str)}
        ext_codes = {k: str(v).lower() if isinstance(v, bool) else str(v) if v is not None else "null"
                     for (k, v) in codes_config.items() if not isinstance(v, str)}
        evaluated = evaluate_file(template_path,
                                  ext_vars=ext_vars,
                                  ext_codes=ext_codes,
                                  jpathdir=os.path.join(os.path.dirname(__file__), "lib"))
        return evaluated

    @staticmethod
    def _parse_multiple_header(template_path, target_path):
        multiple_file_output = False
        multiple_file_dir = None
        with open(template_path, 'r', encoding="utf-8") as template:
            first_line = template.readline()
            if first_line.startswith('// multiple-file-output'):
                multiple_file_output = True
                multiple_file_dir = os.path.dirname(target_path)
            first_line_splitted = first_line.rsplit(":", 1)
            if len(first_line_splitted) > 1:
                multiple_file_dir = first_line_splitted[-1].strip()
        return multiple_file_output, multiple_file_dir



# -*- coding: utf-8 -*-
from marshmallow import fields

from ddb.feature.schema import FeatureSchema


class GitSchema(FeatureSchema):
    """
    Git schema.
    """
    fix_files_permissions = fields.Bool(default=True)



from collections import defaultdict

from ddb.feature.docker.lib.compose.config.types import ServicePort


def apply_resolve_ports_conflicts(compose):
    """
    Resolve ports conflicts in docker compose configuration.
    """
    published = defaultdict(set)

    if 'services' in compose:  # pylint:disable=too-many-nested-blocks
        for service in compose['services'].values():
            if 'ports' in service:
                service_ports = service['ports']
                for index, port in enumerate(tuple(service_ports)):
                    parsed_ports = ServicePort.parse(port)
                    if len(parsed_ports) == 1:
                        parsed_port = parsed_ports[0]
                        new_published = parsed_port.published
                        while new_published in published[parsed_port.protocol]:
                            new_published = new_published + 1
                        published[parsed_port.protocol].add(new_published)
                        if parsed_port.published != new_published:
                            if isinstance(port, str):
                                service_port_data = dict(parsed_port._asdict())
                                service_port_data['published'] = new_published
                                new_port = ServicePort(**service_port_data)
                                port = new_port.legacy_repr()
                                if not new_port.protocol and port.endswith('/tcp'):
                                    port = port[:-4]
                            else:
                                port['published'] = str(new_published)
                            service_ports[index] = port

    return compose



# -*- coding: utf-8 -*-
"""
Writer module
"""

from os.path import normpath, expanduser, expandvars, join

from .checksums import validate_checksum
from .crypto import convert_pem_to_der


def _write_file(path, binary, destination=None):
    if destination:
        path = join(normpath(expandvars(expanduser(destination))), path)
    with open(path, 'wb') as stream:
        stream.write(binary)
    return path


def get_certs_path(output, conf=None, destination=None):
    """
    Get paths of certificate files generated by this module
    """
    if not conf:
        conf = {}

    filenames = conf.get('filenames', {})
    if not destination:
        destination = conf.get('destination')

    private_key_path = filenames.get('private_key', '%s.key.pem') % output
    certificate_path = filenames.get('certificate', '%s.pem') % output

    if destination:
        private_key_path = join(normpath(expandvars(expanduser(destination))), private_key_path)
        certificate_path = join(normpath(expandvars(expanduser(destination))), certificate_path)

    return certificate_path, private_key_path


def write_files(response, output, der, csr, conf=None, destination=None, append_ca_certificate=False, client=None,
                verify_checksum=True):
    """
    Write files contained in response.
    :param response:
    :param output:
    :type output: str
    :param der:
    :param csr:
    :param conf:
    :param append_ca_certificate:
    :param destination:
    :param client:
    """
    # pylint: disable=too-many-locals,too-many-branches
    if not conf:
        conf = {}

    certificate_der = None
    certificate_request_der = None

    should_verify_certificate_der = True

    filenames = conf.get('filenames', {})
    if not destination:
        destination = conf.get('destination')

    generated = {}
    if 'private_key' in response:
        private_key = response['private_key'].encode('ascii')
        generated['private_key'] = \
            _write_file(filenames.get('private_key', '%s.key.pem') % output, private_key, destination=destination)

    if 'certificate' in response:
        certificate = response['certificate'].encode('ascii')
        certificate_der = convert_pem_to_der('certificate', certificate)
        if verify_checksum:
            validate_checksum('certificate', certificate_der, response['sums']['certificate'], True)

        if append_ca_certificate and client:
            info = client.info('')

            ca_cert = info['certificate'] + '\n'

            certificate += ca_cert.encode('ascii')
            certificate_der = convert_pem_to_der('certificate', certificate)
            should_verify_certificate_der = False

        generated['certificate'] = \
            _write_file(filenames.get('certificate', '%s.pem') % output, certificate, destination=destination)
    if csr and 'certificate_request' in response:
        certificate_request_der = convert_pem_to_der('certificate_request',
                                                     response['certificate_request'].encode('ascii'))
        if verify_checksum:
            validate_checksum('certificate_request', certificate_request_der,
                              response['sums']['certificate_request'], True)
        generated['certificate_request'] = \
            _write_file(filenames.get('certificate_request', '%s.csr.pem') % output,
                        response['certificate_request'].encode('ascii'),
                        destination=destination)

    if der:
        if 'certificate' in response:
            generated['certificate_der'] = \
                _write_file(filenames.get('certificate_der', '%s.der') % output,
                            certificate_der,
                            destination=destination)
            if verify_checksum and should_verify_certificate_der:
                with open(filenames.get('certificate_der', '%s.der') % output, 'rb') as der_file:
                    content = der_file.read()
                    validate_checksum('certificate', content, response['sums']['certificate'], True)
        if csr and 'certificate_request' in response:
            generated['certificate_request_der'] = \
                _write_file(filenames.get('certificate_request_der', '%s.csr.der') % output,
                            certificate_request_der,
                            destination=destination)
            if verify_checksum:
                with open(filenames.get('certificate_request_der', '%s.csr.der') % output, 'rb') as der_file:
                    content = der_file.read()
                    validate_checksum('certificate_request', content, response['sums']['certificate_request'], True)
    return generated


def write_stdout(response, der, csr, append_ca_certificate, client=None):
    """
    Writes certificates to stdout.
    :param response:
    :param der:
    :param csr:
    :param append_ca_certificate:
    :param client:
    :return:
    """
    if 'private_key' in response:
        print(response['private_key'])
    if 'certificate' in response:
        print(response['certificate'])

        if append_ca_certificate and client:
            info = client.info('')
            ca_cert = info['certificate'] + '\n'
            print(ca_cert)

    if 'certificate_request' in response:
        print(response['certificate_request'])

    if der:
        if 'certificate' in response:
            print(convert_pem_to_der('certificate', response['certificate']))
        if csr and 'certificate_request' in response:
            print(convert_pem_to_der('certificate_request', response['certificate_request']))



# -*- coding: utf-8 -*-
from typing import ClassVar, Iterable

from ddb.action import Action
from ddb.config import config
from ddb.feature import Feature
from dotty_dict import Dotty

from .actions import UpdateGitignoreAction
from .schema import GitignoreSchema


class GitignoreFeature(Feature):
    """
    Update .gitignore file when a file is generated.
    """

    @property
    def name(self) -> str:
        return "gitignore"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core"]

    @property
    def schema(self) -> ClassVar[GitignoreSchema]:
        return GitignoreSchema

    @property
    def actions(self) -> Iterable[Action]:
        return (
            UpdateGitignoreAction(),
        )

    def _configure_defaults(self, feature_config: Dotty):
        self._configure_defaults_disabled(feature_config)

    @staticmethod
    def _configure_defaults_disabled(feature_config):
        if feature_config.get('disabled') is None:
            feature_config['disabled'] = False
            # Feature is disabled on non-dev environment.
            current = config.data.get('core.env.current')
            available = config.data.get('core.env.available')
            if current and available:
                feature_config['disabled'] = current != available[-1]



# -*- coding: utf-8 -*-
from marshmallow import fields

from ddb.feature.schema import FeatureSchema


class PermissionsSchema(FeatureSchema):
    """
    Permissions Schema
    """
    specs = fields.Dict(fields.String(required=True), fields.String(required=True), allow_none=True, default=None)



# -*- coding: utf-8 -*-
import os
import re
import tempfile
from typing import Union, Iterable, Tuple, Optional

import yaml

from ddb.action.action import AbstractTemplateAction
from ddb.config import config, migrations
from ddb.config.migrations import AbstractPropertyMigration
from ddb.utils.file import TemplateFinder, SingleTemporaryFile
from ddb.utils.process import run


class YttAction(AbstractTemplateAction):
    """
    Render ytt files based on filename suffixes.
    """

    @property
    def name(self) -> str:
        return "ytt:render"

    def _build_template_finder(self):
        return TemplateFinder(config.data.get("ytt.includes"),
                              config.data.get("ytt.excludes"),
                              config.data.get("ytt.suffixes"))

    @staticmethod
    def _escape_config(input_config: dict, with_config: bool = True):
        new = {}
        keywords = config.data["ytt.keywords"]
        keywords_escape_format = config.data["ytt.keywords_escape_format"]
        for key, value in input_config.items():
            if isinstance(value, dict):
                value = YttAction._escape_config(value, False)
            if key in keywords:
                escaped_k = keywords_escape_format % (key,)
                if escaped_k not in value.keys():
                    new[escaped_k] = value
            new[key] = value
        if with_config:
            new['_config'] = {}
            new['_config']['eject'] = config.eject
            new['_config']['args'] = dict(vars(config.args))
            new['_config']['unknown_args'] = list(config.unknown_args)
        return new

    def _render_template(self, template: str, target: str) -> Iterable[Tuple[Union[str, bytes, bool], str]]:
        yaml_config = yaml.safe_dump(YttAction._escape_config(config.data.raw()))

        includes = TemplateFinder.build_default_includes_from_suffixes(
            config.data["ytt.depends_suffixes"],
            config.data["ytt.extensions"]
        )
        template_finder = TemplateFinder(includes, [], config.data["ytt.depends_suffixes"],
                                         os.path.dirname(target),
                                         recursive=False, skip_processed_targets=False)

        depends_files = [template[0] for template in template_finder.items]
        if target in depends_files:
            depends_files.remove(target)

        input_files_args = []
        # pylint:disable=consider-using-with
        yaml_config_file = tempfile.NamedTemporaryFile("w", suffix=".yml", encoding="utf-8", delete=False)

        try:
            try:
                input_files = [template, yaml_config_file.name] + depends_files
                for input_file in input_files:
                    input_files_args += ["-f", input_file]

                yaml_config_file.write("#@data/values")
                yaml_config_file.write(os.linesep)
                yaml_config_file.write("---")
                yaml_config_file.write(os.linesep)
                yaml_config_file.write(yaml_config)
                yaml_config_file.flush()
            finally:
                yaml_config_file.close()

            rendered = run("ytt", *input_files_args, *config.data["ytt.args"])

            yield rendered, target
        finally:
            os.unlink(yaml_config_file.name)

    def _autofix_render_error(self,
                              template: str,
                              target: str,
                              original_template: str,
                              render_error: Exception) -> Optional[str]:
        if not hasattr(render_error, "stderr"):
            return None
        error_message = str(render_error.stderr, encoding="utf-8")

        property_migration_set = set()

        error_match = re.search(r"struct has no \.(.*?)\s+field or method", error_message)
        if error_match:
            property_contains = f"{error_match.group(1)}"

            for property_migration in migrations.get_history():
                if isinstance(property_migration, AbstractPropertyMigration) \
                        and not property_migration.requires_value_migration \
                        and property_contains in property_migration.old_config_key:
                    property_migration_set.add(property_migration)

        if property_migration_set:
            with open(template, "r", encoding="utf-8") as template_file:
                initial_template_data = template_file.read()
                template_data = initial_template_data

            for property_migration in property_migration_set:
                if property_migration and not property_migration.requires_value_migration:
                    property_migration.warn(original_template)

                    template_data = re.sub(r"(#@.*\s+)" +
                                           re.escape("data.values." + property_migration.old_config_key) +
                                           r"(.*?\s*.*?)",
                                           r"\1" + "data.values." + property_migration.new_config_key + r"\2",
                                           template_data)

            if initial_template_data != template_data:
                with SingleTemporaryFile("ddb", "migration", "ytt",
                                         mode="w",
                                         prefix="",
                                         suffix="." + os.path.basename(original_template),
                                         encoding="utf-8") as tmp_file:
                    tmp_file.write(template_data)
                    return tmp_file.name

        return None



import os
import shutil

import pytest
from git import Repo

from ddb.__main__ import load_registered_features
from ddb.config import config
from ddb.feature import features
from ddb.feature.core import CoreFeature
from ddb.feature.git import GitFeature, FixFilePermissionsAction


def test_update_files_disabled(project_loader):
    project_loader("disabled")
    features.register(CoreFeature())
    features.register(GitFeature())

    load_registered_features()

    config.data['git.fix_file_permissions'] = False

    action = FixFilePermissionsAction()
    assert action.disabled is True


@pytest.mark.skipif("os.name == 'nt'")
class TestGitFixFilesPermissionsAction:
    def test_update_files_simple(self, project_loader):
        project_loader("simple")
        features.register(CoreFeature())
        features.register(GitFeature())

        load_registered_features()

        repo = Repo.init(config.path.project_home)
        repo.git.add('.')
        repo.git.update_index('.gitignore', chmod='+x')
        repo.git.commit('-m "Initial commit"')

        action = FixFilePermissionsAction()
        action.execute()

        assert os.access(os.path.join(config.path.project_home, '.gitignore'), os.X_OK)

        shutil.rmtree(config.path.project_home)

    def test_update_files_with_submodule(self, project_loader):
        project_loader("with_submodules")
        features.register(CoreFeature())
        features.register(GitFeature())

        load_registered_features()

        path_main = config.path.project_home
        path_submodule = os.path.join(path_main, 'submodule')

        submodule = Repo.init(path_submodule)
        submodule.create_remote('origin', 'https://github.com/inetum-orleans/docker-devbox-ddb')
        submodule.git.add('.')
        submodule.git.update_index('.gitignore', chmod='+x')
        submodule.git.commit('-m "Initial commit"')

        repo = Repo.init(path_main)
        repo.git.add('.')
        repo.create_submodule('a-sub-module', 'submodule', no_checkout=True)
        repo.git.update_index('.gitignore', chmod='+x')
        repo.git.commit('-m "Initial commit"')

        action = FixFilePermissionsAction()
        action.execute()

        assert os.access(os.path.join(config.path.project_home, '.gitignore'), os.X_OK)
        assert os.access(os.path.join(config.path.project_home, 'submodule', '.gitignore'), os.X_OK)
        assert not os.access(os.path.join(config.path.project_home, '.gitmodules'), os.X_OK)



# -*- coding: utf-8 -*-
from abc import ABC
from argparse import ArgumentParser
from typing import Any, Callable, Union

from ddb.cache import project_cache
from ddb.context import context
from ddb.event import bus
from ddb.registry import RegistryObject, DefaultRegistryObject


class Phase(RegistryObject, ABC):  # pylint:disable=abstract-method
    """
    Step into a lifecycle flow.
    """

    def configure_parser(self, parser: ArgumentParser):
        """
        Configure the argument parser.
        """

    def execute(self):
        """
        Execute the phase.
        """

    @property
    def allow_unknown_args(self):
        """
        Check if this phase allow unkown args.
        """
        return False


class DefaultPhase(DefaultRegistryObject, Phase):
    """
    Default implementation for a phase.
    """

    def __init__(self,
                 name: str,
                 description: Union[str, None] = None,
                 parser: Callable[[ArgumentParser], Any] = None,
                 run_once: bool = False,
                 allow_unknown_args: bool = False):
        super().__init__(name, description)
        self.run_once = run_once
        self._parser = parser
        self._allow_unknown_args = allow_unknown_args

    @property
    def allow_unknown_args(self):
        return self._allow_unknown_args

    def configure_parser(self, parser: ArgumentParser):
        if self._parser:
            self._parser(parser)

    def execute(self):
        cache = project_cache()
        runned_cache_key = "phase.runned." + self.name
        if self.run_once and cache.get(runned_cache_key):
            return
        bus.emit("phase:" + self.name)
        cache.set(runned_cache_key, True)
        cache.flush()


def execute_phase(phase: Phase):
    """
    Execute a phase with context update.
    """
    context.phase = phase
    try:
        phase.execute()
    finally:
        context.phase = None



import os
from subprocess import CalledProcessError
from typing import Union, Iterable, Callable

from ddb.action.action import EventBinding, InitializableAction
from ddb.command import commands
from ddb.event import events
from ddb.utils.file import write_if_different
from ddb.utils.process import run


def is_smartcd_installed():
    """
    Run smartcd command to check if it"s installed.
    """

    try:
        run("bash", "-ic", "smartcd", "--help")
        return True
    except CalledProcessError:
        return False


class SmartcdAction(InitializableAction):
    """
    Generate bash_enter/bash_leave files if smartcd is installed
    """

    @property
    def event_bindings(self) -> Union[Iterable[Union[Callable, str, EventBinding]], Union[Callable, str, EventBinding]]:
        return events.phase.configure

    @property
    def name(self) -> str:
        return "smartcd:generate"

    @property
    def disabled(self) -> bool:
        return os.name == 'nt'

    @staticmethod
    def execute():
        """
        Write .bash_enter/.bash_leave files.
        :return:
        """
        if not commands.has("activate") and not commands.has("deactivate"):
            return
        if not is_smartcd_installed():
            return

        bash_enter_content = [
            "echo [smartcd] Activate",
            "$(ddb activate)"
        ]
        bash_enter = '\n'.join(bash_enter_content) + '\n'

        bash_leave_content = [
            "echo [smartcd] Deactivate",
            "$(ddb deactivate)"
        ]
        bash_leave = '\n'.join(bash_leave_content) + '\n'

        if write_if_different(".bash_enter", bash_enter):
            events.file.generated(source=None, target=".bash_enter")

        if write_if_different(".bash_leave", bash_leave):
            events.file.generated(source=None, target=".bash_leave")


class WindowsProjectActivate(InitializableAction):
    """
    Generate ddb_activate.bat/ddb_deactivate.bat for windows environement
    TODO find and activate something equivalent to smartcd but for windows ?
    """

    @property
    def event_bindings(self) -> Union[Iterable[Union[Callable, str, EventBinding]], Union[Callable, str, EventBinding]]:
        return events.phase.configure

    @property
    def name(self) -> str:
        return "smartcd:windows:generate-project-activate"

    @property
    def disabled(self) -> bool:
        return os.name != 'nt'

    @staticmethod
    def execute():
        """
        Write .bash_enter/.bash_leave files.
        :return:
        """
        if not commands.has("activate") and not commands.has("deactivate"):
            return

        commands_activate = ["@echo off", "REM ddb:shim", "SHIFT", "set command=(ddb activate)", "%command%>cmd.txt",
                             "set /p execution=<cmd.txt", "del cmd.txt", "%execution% \"%*\""]
        command_activate = '\n'.join(commands_activate) + '\n'

        if write_if_different("ddb_activate.bat", command_activate):
            events.file.generated(source=None, target="ddb_activate.bat")

        commands_deactivate = ["@echo off", "REM ddb:shim", "SHIFT", "set command=(ddb deactivate)",
                               "%command%>cmd.txt", "set /p execution=<cmd.txt", "del cmd.txt", "%execution% \"%*\""]
        command_deactivate = '\n'.join(commands_deactivate) + '\n'

        if write_if_different("ddb_deactivate.bat", command_deactivate):
            events.file.generated(source=None, target="ddb_deactivate.bat")
